import importlib
import os
import pathlib
import platform
import re
import sys
import time
import types
import typing

import torch
import torchvision

from .._version import __version__

T = typing.TypeVar("T")
T_Any = typing.TypeVar("T_Any", bound=typing.Any)


def get_env_info_str() -> str:
    """
    Return a string with the Helios header and the environment information.

    Returns:
        The message string.
    """
    msg = r"""
#===========================================================================#
          _______  _       _________ _______  _______
|\     /|(  ____ \( \      \__   __/(  ___  )(  ____ \
| )   ( || (    \/| (         ) (   | (   ) || (    \/
| (___) || (__    | |         | |   | |   | || (_____
|  ___  ||  __)   | |         | |   | |   | |(_____  )
| (   ) || (      | |         | |   | |   | |      ) |
| )   ( || (____/\| (____/\___) (___| (___) |/\____) |
|/     \|(_______/(_______/\_______/(_______)\_______)
    """
    msg += (
        "\nEnvironment info: "
        f"\n\tHelios: {__version__}"
        f"\n\tPyTorch: {torch.__version__}"
        f"\n\tTorchVision: {torchvision.__version__}"
        f"\n\tOS: {platform.platform()}"
        f"\n\tPython: {sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[3]}"
    )
    if torch.cuda.is_available():
        msg += f"\n\tCUDA version: {torch.version.cuda}"
    msg += "\n"
    msg += "#===========================================================================#"
    msg += "\n\n"
    return msg


def get_from_optional(opt_var: T | None, raise_on_empty: bool = False) -> T:
    """
    Ensure the given variable is not :code:`None` and return it.

    This is useful when dealing with variables that can be ``None`` at declaration but
    are set elsewhere. In those instances, mypy is unable to determine that the variable
    was set, so it will issue a warning. The workaround is to add asserts, but that can
    get tedious very quickly. This function can be used as an alternative.

    Example:
        .. code-block:: python

            var: int | None = None
            # ... Set var to a valid value some place else.

            assert var is not None
            v = var

            # Alternatively:
            v = core.get_from_optional(var)


    Args:
        opt_var: the optional variable.
        raise_on_empty: if True, an exception is raised when the optional is ``None``.

    Returns:
        The variable without the optional.

    Raises:
          RuntimeError: if the ``opt_var`` is ``None`` and ``raise_on_empty`` is true.

    """
    if not raise_on_empty:
        assert opt_var is not None
    else:
        raise RuntimeError("error: optional cannot be empty")
    return opt_var


def convert_to_list(var: T | list[T] | tuple[T, ...]) -> list[T]:
    """
    Convert the input into a list if it's not one already.

    Example:
        .. code-block:: python

            def some_fun(x: int | list[int]) -> None:
                if isinstance(x, list):
                    x = [x]
                for elem in x:
                    ...

                # The above code an be replaced with this:
                for elem in convert_to_list(x):
                    ...

    Args:
        var: an object that can be either a single object or a list.

    Returns:
        If the input was a list, no operation is done. Otherwise, the object is converted
        to a list and returned.
    """
    if isinstance(var, list):
        return var
    if isinstance(var, tuple):
        return list(var)
    return [var]


class ChdirContext:
    """
    Allow switching between the current working directory and another within a scope.

    The intention is to facilitate temporary switches of the current working directory
    (such as when attempting to resolve relative paths) by creating a context in which the
    working directory is automatically switched to a new one. Upon exiting of the context,
    the original working directory is restored.

    Example:
        .. code-block:: python

            os.chdir(".")   # <- Starting working directory
            with ChdirContext("/new/path") as prev_cwd:
                # prev_cwd is the starting working directory
                Path.cwd() # <- This is /new/path now
                ...
            Path.cwd() # <- Back to the starting working directory.

    Args:
        target_path: the path to switch to.
    """

    def __init__(self, target_path: pathlib.Path):
        """
        Create the context manager with the given path.

        Args:
            target_path: the path to switch to.
        """
        self.start_path = pathlib.Path.cwd()
        self.target_path = target_path

    def __enter__(self) -> pathlib.Path:
        """
        Perform the switch from the current working directory to the new one.

        Returns:
            The previous working directory.
        """
        os.chdir(self.target_path)
        return self.start_path

    def __exit__(
        self,
        exc_type: type[Exception] | None,
        exc_value: Exception | None,
        exc_traceback: types.TracebackType | None,
    ) -> None:
        """Restores the previous working directory."""
        os.chdir(self.start_path)


class AverageTimer:
    """
    Compute elapsed times using moving average.

    The timer will determine the elapsed time between a series of points using a sliding
    window moving average.

    Args:
        sliding_window: number of steps over which the moving average will be computed.
    """

    def __init__(self, sliding_window: int = 200):
        """
        Create the timer with the given sliding window.

        Args:
            sliding_window: steps over which the moving average will be computed.
        """
        self._sliding_window = sliding_window

        self._time_sum: float = 0
        self._step_count: int = 0
        self._current_time: float = 0
        self._avg_time: float = 0
        self.start()

    def start(self) -> None:
        """Start the timer."""
        self._current_time = time.time()

    def record(self) -> None:
        """Record a new step in the timer."""
        self._step_count += 1
        self._time_sum += time.time() - self._current_time
        self._avg_time = self._time_sum / self._step_count

        if self._step_count > self._sliding_window:
            self._step_count = 0
            self._time_sum = 0

        self._current_time = time.time()

    def get_average_time(self) -> float:
        """Return the moving average over the current step count."""
        return self._avg_time


class Registry:
    """
    Provides a name to object mapping to allow users to create custom types.

    Example:
        .. code-block:: python

            # Create a registry:
            TEST_REGISTRY = Registry("test")

            # Register as a decorator:
            @TEST_REGISTRY.register
            class TestClass:
                ...

            # Register in code:
            TEST_REGISTRY.register(TestClass)
            TEST_REGISTRY.register(test_function)

    Args:
        name: the name of the registry.
    """

    def __init__(self, name: str):
        """
        Create the registry with the given name.

        Args:
            name: the name of the registry.
        """
        self._name = name
        self._obj_map: dict[str, typing.Any] = {}

    def _do_register(self, name: str, obj: typing.Any, suffix: str | None = None) -> None:
        """
        Register the function/class.

        Args:
            name: the name of the object to register.
            obj: the object to register.
            suffix: (optional) suffix to add to the name upon registration.
        """
        if isinstance(suffix, str):
            name = name + "_" + suffix

        assert (
            name not in self._obj_map
        ), f"error: an object named '{name}' already exists in the "
        f"'{self._name}' registry"

        self._obj_map[name] = obj

    def register(self, obj: T_Any, suffix: str | None = None) -> T_Any:
        """
        Register the given object.

        Args:
            obj: the type to add. Must have a __name__ attribute.
            suffix: (optional) the suffix to add to the type name.

        Returns:
            The registered type.
        """
        name = obj.__name__
        self._do_register(name, obj, suffix)
        return obj

    def get(self, name: str, suffix: str | None = None) -> typing.Any:
        """
        Get the object that corresponds to the given name.

        Args:
            name: the name of the type.
            suffix: (optional) the suffix to use if the type isn't found with the given
                name.

        Returns:
            The requested type.

        Raises:
            KeyError: if no object with the given name is found in the registry.
        """
        ret = self._obj_map.get(name)
        if ret is None and suffix is not None:
            name_suff = name + "_" + suffix
            ret = self._obj_map.get(name_suff)
            print(f"warning: found {name_suff} instead of {name}")
            if ret is None:
                raise KeyError(
                    f"No object called '{name}' found in the '{self._name}' registrar"
                )
        elif ret is None:
            raise KeyError(
                f"No object called '{name}' found in the '{self._name}' registrar"
            )
        return ret

    def __contains__(self, name: str) -> bool:
        """
        Check if the registry contains the given name.

        Args:
            name: the name to check.

        Returns:
            True if the name exists, false otherwise.
        """
        return name in self._obj_map

    def __iter__(self) -> typing.Iterable:
        """Get an iterable over the registry items."""
        return iter(self._obj_map.items())

    def __str__(self) -> str:
        """Get the name of the registry."""
        return self._name

    def keys(self) -> typing.Iterable:
        """
        Return a set-like object providing a view into the registry's keys.

        Return:
            An iterable of the registry keys.
        """
        return self._obj_map.keys()


def update_all_registries(
    root: pathlib.Path, recurse: bool = True, import_prefix: str = ""
) -> None:
    """
    Ensure all registered types get added to their corresponding registries.

    This function serves as a way of automatically registering all types into their
    corresponding registries within a package. Normally, you'd have to manually include
    each module that contains a registered type to ensure that it gets registered. This
    can easily cascade if modules are nested inside packages, whereby the top-level module
    has to (somehow) ensure that all child modules get imported to ensure everything works
    correctly.

    This function offers an alternative, whereby it will automatically scan all
    modules and sub-packages within a given package and import only those files that
    register a type. To do this, there are a few assumptions:

    #. Each package MUST contain an ``__init__.py`` (namespace packages are not
       supported)
    #. A module is included if and only if there is at least one line that contains the
       following pattern: ``@<any non-whitespace character(s)>.register``.

    Example:
        Suppose we have a project with the following structure:

        .. code-block:: text

            main.py
            my_package/
            |---__init__.py
            |---some_class.py <- This registers a type.
            |---some_funcs.py <- Doesn't register anything.
            |---sub_package/
            |   |---__init__.py
            |   |---another_type.py <- Registers
            |   |---another_func.py <- Doesn't register.

        We can then do the following inside ``main.py``:

        .. code-block:: python

            import helios.core as hlc
            ...
            hlc.update_all_registries(Path.cwd() / "my_package", recurse=True)

        The function will recursively walk through ``my_package`` and import the
        following:

        * ``my_package.some_class``
        * ``my_package.sub_package.another_type``

        After the function returns, the corresponding registries will have been populated
        with the types and they can be used elsewhere in the code.

    Args:
        root: the path to the root package.
        recurse: if True, recursively search through sub-packages. Defaults to true.
        import_prefix: (optional) prefix to be added when imported. Defaults to empty.

    Raises:
        RuntimeError: if the given path isn't a valid directory or if the directory is not
            Python package with ``__init__.py``.
    """
    if not root.is_dir():
        raise RuntimeError(f"error: expected {str(root)} to be a valid directory")

    if import_prefix == "":
        import_prefix = root.stem
    else:
        import_prefix += f".{root.stem}"

    # Ensure the __init__.py exists
    init_path = root / "__init__.py"
    if not init_path.exists():
        raise RuntimeError(f"error: {str(root)} is not a Python package")

    modules: list[tuple[pathlib.Path, str]] = []
    for path in root.iterdir():
        stem = path.stem
        if path.is_dir() and recurse:
            if stem.startswith(("__", ".")):
                continue
            update_all_registries(path, True, import_prefix)

        if path.is_file() and path.suffix == ".py" and stem != "__init__":
            modules.append((path, import_prefix + f".{path.stem}"))

    import_modules: list[str] = []
    p = re.compile(r"@.+\.register\s")
    for path, tag in modules:
        with path.open("r", encoding="utf-8") as infile:
            lines = infile.readlines()
        for line in lines:
            if p.match(line):
                import_modules.append(tag)
                break

    for module in import_modules:
        importlib.import_module(module)


def safe_torch_load(
    f: str | os.PathLike | typing.BinaryIO | typing.IO[bytes],
    **kwargs: typing.Any,
) -> typing.Any:
    """
    Wrap :code:`torch.load` to handle safe loading.

    This function will automatically set :code:`weights_only` to true when calling
    ``torch.load``. You are encouraged to use this function instead of the plain
    :code:`torch.load` to ensure safe loading.

    .. warning::
        :code:`weights_only` is set automatically by this function. **do not** set this
        value yourself when using this function.

    args:
        f: a file-like object (has to implement ``read()``, ``readline()``, ``tell()``,
            and ``seek()``), or a string or a ``os.pathlike`` object containing a file
            name.
        **kwargs: keyword arguments to pass to :code:`torch.load`.

    Returns:
        The result of calling :code:`torch.load`.
    """
    return torch.load(f, **kwargs, weights_only=True)



import copy
import typing

import torch
from torch import nn


class EMA(nn.Module):
    """
    Implements Exponential Moving Average (EMA).

    Args:
        net: the bare network on which EMA will be performed.
        decay: decay rate. Defaults to 0.9997.
        device: (optional) the device to be used.
    """

    def __init__(
        self, net: nn.Module, decay: float = 0.9997, device: torch.device | None = None
    ):
        """Create the EMA wrapper."""
        super().__init__()

        self._module = copy.deepcopy(net)
        self._module = self._module.eval()
        self._decay = decay
        self._device = device

        if self._device is not None:
            self._module.to(device=device)

    @property
    def module(self) -> nn.Module:
        """Get the underlying network."""
        return self._module

    @torch.no_grad()
    def _update(self, net: nn.Module, update_fn: typing.Callable) -> None:
        for ema_v, net_v in zip(
            self._module.state_dict().values(),
            net.state_dict().values(),
            strict=True,
        ):
            if self._device:
                net_v = net_v.to(device=self._device)
            ema_v.copy_(update_fn(ema_v, net_v))

    def update(self, net: nn.Module) -> None:
        """Update the weights using EMA from the given network."""
        self._update(
            net, update_fn=lambda e, m: self._decay * e + (1.0 - self._decay) * m
        )

    def set(self, net: nn.Module) -> None:
        """Re-set the base weights."""
        self._update(net, update_fn=lambda e, m: m)

    def forward(self, *args: typing.Any, **kwargs: typing.Any):
        """
        Evaluate the EMA wrapper on the network inputs.

        Args:
            args: named parameters for your network's forward function.
            kwargs: keyword arguments for your network's forward function.
        """
        return self._module(*args, **kwargs)



"""
Optuna module for Helios.

This contains the classes and utilities related to the Optuna integration.

.. warning::
    This module **requires** that Optuna be installed.
"""

import importlib.util

if importlib.util.find_spec("optuna") is None:
    raise ImportError("error: Optuna is required to access this module")

from .plugin import OptunaPlugin
from .utils import (
    StudyArgs,
    checkpoint_sampler,
    create_or_load_study,
    create_study_starting_from_trial,
    restore_sampler,
)

__all__ = [
    "OptunaPlugin",
    "StudyArgs",
    "checkpoint_sampler",
    "create_or_load_study",
    "create_study_starting_from_trial",
    "restore_sampler",
]



import dataclasses as dc
import math
import pathlib
import pickle
import random
import sys
import time
import typing

import numpy as np
import numpy.typing as npt
import pytest
import torch

from helios import core
from helios.core import cuda, rng


@dc.dataclass
class SafeType:
    a: int
    b: str


class TestUtils:
    def test_get_from_optional(self) -> None:
        opt: int | None = None

        with pytest.raises(RuntimeError):
            core.get_from_optional(opt, raise_on_empty=True)

        opt = 1
        assert core.get_from_optional(opt) == 1

    def test_convert_to_list(self) -> None:
        value = 0
        x: int | list[int] | tuple[int, ...] = value

        res = core.convert_to_list(x)
        assert isinstance(res, list)
        assert len(res) == 1
        assert res[0] == value

        x = [value]
        res = core.convert_to_list(x)
        assert res == x

        x = (1, 3, 4)
        res = core.convert_to_list(x)
        assert len(x) == len(res)
        assert list(x) == res

    def test_chdir_context(self, tmp_path: pathlib.Path) -> None:
        cur_dir = pathlib.Path.cwd()
        with core.ChdirContext(tmp_path) as cwd:
            assert cwd == cur_dir
            assert pathlib.Path.cwd() == tmp_path

        assert pathlib.Path.cwd() == cur_dir

    def test_average_timer(self) -> None:
        timer = core.AverageTimer()

        timer.start()
        for _ in range(10):
            time.sleep(0.1)
            timer.record()

        avg_time = timer.get_average_time()
        assert math.isclose(avg_time, 0.1, abs_tol=1e-2)

    def test_register(self) -> None:
        test_registry = core.Registry("test")

        test_registry.register(sample_fun)

        assert str(test_registry) == "test"
        assert len(typing.cast(typing.Sized, test_registry.keys())) == 1
        assert "sample_fun" in test_registry
        assert test_registry.get("sample_fun") == sample_fun

        with pytest.raises(KeyError):
            test_registry.get("foo")

    def test_update_registries(self) -> None:
        root = pathlib.Path(__file__)
        root = root.parent
        # HACK: Ensure that the current path gets added to sys path, otherwise imports
        # won't work.
        sys.path.append(str(root))
        core.update_all_registries(root / "registry_test", recurse=True)

        import registry_test as rt  # type: ignore[import-not-found]

        assert len(rt.FUNC_REGISTRY.keys()) == 2
        assert all(var in rt.FUNC_REGISTRY for var in ("foo", "bar"))

    def test_safe_torch_load(self, tmp_path: pathlib.Path) -> None:
        out_file = tmp_path / "safe.pth"
        base_dict = {"a": 1, "b": "foo", "c": SafeType(2, "bar")}
        torch.save(base_dict, out_file)

        # Loading without registering the type should fail.
        with pytest.raises(pickle.UnpicklingError):
            core.safe_torch_load(out_file)

        # Now register the type and loading should pass.
        torch.serialization.clear_safe_globals()
        torch.serialization.add_safe_globals([SafeType])
        ret_dict = core.safe_torch_load(out_file)
        assert base_dict == ret_dict
        torch.serialization.clear_safe_globals()


def sample_fun() -> int:
    return 1


class TestCUDA:
    def test_functions(self) -> None:
        if torch.cuda.is_available():
            cuda.requires_cuda_support()
        else:
            with pytest.raises(RuntimeError):
                cuda.requires_cuda_support()

    def _check_cudnn(self, val: bool) -> None:
        torch.backends.cudnn.benchmark = val
        assert torch.backends.cudnn.benchmark == val
        with cuda.DisableCuDNNBenchmarkContext():
            assert not torch.backends.cudnn.benchmark

        assert torch.backends.cudnn.benchmark == val

    def test_disable_cudnn_context(self) -> None:
        if torch.cuda.is_available():
            self._check_cudnn(True)
            self._check_cudnn(False)


@dc.dataclass
class ExpectedRNG:
    torch_vals = torch.tensor([0, 1, 8, 6, 5, 7, 3, 9, 6, 9])
    rand_vals = [0, 9, 2, 3, 6, 6, 2, 6, 7, 4]
    np_vals = [3, 2, 3, 4, 4, 3, 1, 3, 3, 1]


class TestRNG:
    def check_torch(self, a: torch.Tensor, b: torch.Tensor) -> None:
        assert torch.all(a == b)

    def check_rand(self, a: list[int], b: list[int]) -> None:
        assert a == b

    def check_np(self, a: list[int], b: npt.NDArray) -> None:
        assert np.all(a == b)

    def test_seed_rngs(self) -> None:
        exp = ExpectedRNG()
        rng.seed_rngs()

        np_gen = rng.get_default_numpy_rng().generator

        self.check_torch(exp.torch_vals, torch.randint(10, [10]))
        self.check_rand(exp.rand_vals, [random.randint(0, 9) for _ in range(10)])
        self.check_np(exp.np_vals, np_gen.integers(0, 10, 10))

    def test_rng_restore(self) -> None:
        exp = ExpectedRNG()
        rng.seed_rngs()
        np_gen = rng.get_default_numpy_rng().generator

        # Check the first 5 entries.
        self.check_torch(exp.torch_vals[:5], torch.randint(10, [5]))
        self.check_rand(exp.rand_vals[:5], [random.randint(0, 9) for _ in range(5)])
        self.check_np(exp.np_vals[:5], np_gen.integers(0, 10, 5))

        # Grab the states and re-seed the generators.
        state = rng.get_rng_state_dict()
        rng.seed_rngs(0)
        np_gen = rng.get_default_numpy_rng().generator

        # Generate a few numbers to move the generators along.
        torch.randint(10, [10])
        [random.randint(0, 9) for _ in range(10)]
        np_gen.integers(0, 10, 10)

        # Now restore the RNG state and read the final 5 numbers.
        rng.load_rng_state_dict(state)
        np_gen = rng.get_default_numpy_rng().generator

        self.check_torch(exp.torch_vals[5:], torch.randint(10, [5]))
        self.check_rand(exp.rand_vals[5:], [random.randint(0, 9) for _ in range(5)])
        self.check_np(exp.np_vals[5:], np_gen.integers(0, 10, 5))



import dataclasses as dc
import functools
import pathlib
import typing

import optuna
import pytest

import helios.model as hlm
import helios.plugins.optuna as hlpo
import helios.trainer as hlt
from helios.core import rng
from helios.plugins.optuna import utils


class PluginModel(hlm.Model):
    def __init__(self, save_name: str):
        super().__init__(save_name)

    def setup(self, fast_init: bool = False) -> None:
        pass


# Ignore the warnings coming from optuna.
@pytest.mark.filterwarnings(
    ("ignore::optuna.exceptions.ExperimentalWarning"), ("ignore::FutureWarning")
)
class TestOptunaPlugin:
    def test_plugin_id(self) -> None:
        assert hasattr(hlpo.OptunaPlugin, "plugin_id")
        assert hlpo.OptunaPlugin.plugin_id == "optuna"

    def test_invalid_storage(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")
            plugin.is_distributed = True
            with pytest.raises(ValueError):
                plugin.setup()

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_configure(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")
            trainer = hlt.Trainer()

            plugin.configure_trainer(trainer)
            assert len(trainer.plugins) == 1
            assert trainer.plugins["optuna"] == plugin
            assert len(trainer.train_exceptions) == 1
            assert trainer.train_exceptions[0] == optuna.TrialPruned

            model = PluginModel("plugin-model")
            plugin.configure_model(model)
            assert model.save_name == "plugin-model_trial-0"

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def check_in_range(
        self, val: typing.Any, t: type, low: typing.Any, high: typing.Any
    ) -> None:
        assert isinstance(val, t)
        assert low <= val <= high

    def check_in_sequence(self, val: typing.Any, t: type, seq: typing.Container) -> None:
        assert isinstance(val, t)
        assert val in seq

    def test_suggest(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")

            with pytest.raises(KeyError):
                plugin.suggest("foo", "bar")

            seq = [1, 2, 3]
            self.check_in_sequence(
                plugin.suggest("categorical", "val1", choices=seq), int, seq
            )

            low = 0
            high = 10
            self.check_in_range(
                plugin.suggest("int", "val2", low=low, high=high), int, low, high
            )

            high = 1
            self.check_in_range(
                plugin.suggest("float", "val3", low=low, high=high), float, low, high
            )

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_state_dict(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = hlpo.OptunaPlugin(trial, "accuracy")
            x = plugin.suggest("float", "x", low=-10, high=10)

            state_dict = plugin.state_dict()
            assert len(state_dict) == 1
            assert "x" in state_dict
            assert state_dict["x"] == x

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)


# Allow accessing private members so we can test things properly.
# ruff: noqa: SLF001
class TestOptunaUtils:
    def test_get_backup_name(self, tmp_path: pathlib.Path) -> None:
        test_entries = 4
        name = utils._get_backup_name(tmp_path, "test-file", utils._BackupType.FILE)
        assert name == "test-file_bkp-0"

        for i in range(test_entries):
            (tmp_path / f"test-file_bkp-{i}.txt").touch()

        name = utils._get_backup_name(tmp_path, "test-file", utils._BackupType.FILE)
        assert name == f"test-file_bkp-{test_entries}"

        # Now test folders
        name = utils._get_backup_name(tmp_path, "test-dir", utils._BackupType.DIR)
        assert name == "test-dir_bkp-0"

        for i in range(test_entries):
            (tmp_path / f"test-dir_bkp-{i}").mkdir()

        name = utils._get_backup_name(tmp_path, "test-dir", utils._BackupType.DIR)
        assert name == f"test-dir_bkp-{test_entries}"

    def test_backup_study(self, tmp_path: pathlib.Path) -> None:
        base_study = tmp_path / "base_study.db"
        base_study.touch()
        assert len(list(tmp_path.glob("*.db"))) == 1

        utils._backup_study(base_study)
        assert len(list(tmp_path.glob("*.db"))) == 2

        bkp_study = tmp_path / "base_study_bkp-0.db"
        assert bkp_study.exists()

    def test_backup_samplers(self, tmp_path: pathlib.Path) -> None:
        test_entries = 4
        base_dir = tmp_path / "base"
        base_dir.mkdir()
        for i in range(test_entries):
            (base_dir / f"sampler_{i}.pkl").touch()

        def key(p: pathlib.Path) -> int:
            return int(p.stem.split("_")[-1])

        base_files = list(base_dir.glob("*.pkl"))
        base_files.sort(key=key)

        utils._backup_sampler_chkpts(base_dir)
        bkp_dir = tmp_path / "base_bkp-0"
        assert bkp_dir.exists()

        bkp_files = list(bkp_dir.glob("*.pkl"))
        bkp_files.sort(key=key)
        assert len(base_files) == len(bkp_files)

        for f, b in zip(base_files, bkp_files, strict=True):
            assert f.name == b.name

    def test_check_study_args(self) -> None:
        args: utils.StudyArgs = {}

        with pytest.raises(RuntimeError):
            utils._check_study_args(**args)

        args["storage"] = None
        with pytest.raises(RuntimeError):
            utils._check_study_args(**args)

        args["storage"] = ""
        with pytest.raises(RuntimeError):
            utils._check_study_args(**args)

        args["load_if_exists"] = False
        with pytest.raises(RuntimeError):
            utils._check_study_args(**args)

        args["load_if_exists"] = True
        args["storage"] = 1  # type: ignore[typeddict-item]
        with pytest.raises(TypeError):
            utils._check_study_args(**args)

        args["storage"] = "some_path.db"
        with pytest.raises(RuntimeError):
            utils._check_study_args(**args)

    def test_get_storage_path(self, tmp_path: pathlib.Path) -> None:
        db = tmp_path / "test.db"
        db.touch()
        args: utils.StudyArgs = {"storage": f"sqlite:///{db}"}

        path = utils._get_storage_path(**args)
        assert path == db

    def test_checkpoint_sampler(self, tmp_path: pathlib.Path) -> None:
        @dc.dataclass
        class TestRun:
            samples: list[float]
            chkpt_root: pathlib.Path

        num_trials = 10
        run1 = TestRun([], tmp_path / "run1")
        run2 = TestRun([], tmp_path / "run2")

        run1.chkpt_root.mkdir(exist_ok=True)
        run2.chkpt_root.mkdir(exist_ok=True)

        def objective(
            trial: optuna.Trial,
            raise_error: bool,
            run: TestRun,
        ) -> float:
            hlpo.checkpoint_sampler(trial, run.chkpt_root)
            res = trial.suggest_float("accuracy", 0, 1)
            if raise_error and trial.number == num_trials // 2:
                raise RuntimeError("half-way stop")
            run.samples.append(res)
            return res

        def create_study(
            sampler: optuna.samplers.BaseSampler | None = None,
        ) -> optuna.Study:
            return optuna.create_study(
                study_name="chkpt_test",
                sampler=optuna.samplers.TPESampler(seed=rng.get_default_seed())
                if sampler is None
                else sampler,
            )

        study = create_study()
        study.optimize(
            functools.partial(
                objective,
                raise_error=False,
                run=run1,
            ),
            n_trials=num_trials,
        )

        chkpts = [chkpt.stem for chkpt in run1.chkpt_root.glob("*.pkl")]
        chkpts.sort()
        assert len(chkpts) == num_trials
        assert all(chkpt == f"sampler_trial-{i}" for i, chkpt in enumerate(chkpts))
        del study

        study = create_study()
        sampler = hlpo.restore_sampler(run2.chkpt_root)
        assert sampler is None

        with pytest.raises(RuntimeError):
            study.optimize(
                functools.partial(
                    objective,
                    raise_error=True,
                    run=run2,
                ),
                n_trials=num_trials,
            )

        sampler = hlpo.restore_sampler(run2.chkpt_root)
        assert sampler is not None
        study = create_study(sampler)
        study.optimize(
            functools.partial(objective, raise_error=False, run=run2),
            n_trials=num_trials // 2,
        )
        assert run2.samples == run1.samples

    def test_create_study(self, tmp_path: pathlib.Path) -> None:
        args: utils.StudyArgs = {}
        args["storage"] = f"sqlite:///{str(tmp_path / 'basic.db')}"
        args["load_if_exists"] = True
        args["study_name"] = "basic"

        sampler_path = tmp_path / "samplers"
        sampler_path.mkdir()

        study = utils.create_or_load_study(sampler_path=sampler_path, **args)

        def objective(trial: optuna.Trial) -> float:
            hlpo.checkpoint_sampler(trial, sampler_path)
            res = trial.suggest_float("accuracy", 0, 1)
            return res

        study.optimize(objective, n_trials=2)

        del study

        # Now check backups.
        utils.create_or_load_study(
            sampler_path=sampler_path, backup_samplers=True, backup_study=True, **args
        )

        dbs = [p.name for p in tmp_path.glob("*.db")]
        assert len(dbs) == 2
        assert "basic_bkp-0.db" in dbs

        sampler_bkp = tmp_path / "samplers_bkp-0"
        assert sampler_bkp.exists()

        base_samplers = [p.name for p in sampler_path.glob("*.pkl")]
        base_samplers.sort()
        bkp_samplers = [p.name for p in sampler_bkp.glob("*.pkl")]
        bkp_samplers.sort()
        assert base_samplers == bkp_samplers

    def test_load_study(self, tmp_path: pathlib.Path) -> None:
        def create_study(
            name: str,
            add_sampler: bool,
        ) -> tuple[optuna.Study, pathlib.Path]:
            args: utils.StudyArgs = {
                "storage": f"sqlite:///{str(tmp_path / f'{name}.db')}",
                "load_if_exists": True,
                "study_name": name,
            }
            if add_sampler:
                args["sampler"] = optuna.samplers.TPESampler(seed=rng.get_default_seed())

            sampler_path = tmp_path / f"{name}_samplers"
            sampler_path.mkdir(exist_ok=True)

            return utils.create_or_load_study(
                sampler_path=sampler_path, **args
            ), sampler_path

        def objective(
            trial: optuna.Trial,
            samples: list[float],
            sampler_path: pathlib.Path,
            raise_at: int | None = None,
        ) -> float:
            utils.checkpoint_sampler(trial, sampler_path)
            res = trial.suggest_float("accuracy", 0, 1)
            if raise_at is not None and trial.number == raise_at:
                raise RuntimeError("stop")
            samples.append(res)
            return res

        base_samples: list[float] = []
        study, base_samplers = create_study("base", True)
        study.optimize(
            functools.partial(
                objective,
                samples=base_samples,
                sampler_path=base_samplers,
            ),
            n_trials=4,
        )
        assert len(base_samples) == 4
        del study

        samples: list[float] = []
        study, samplers = create_study("resume", True)
        with pytest.raises(RuntimeError):
            study.optimize(
                functools.partial(
                    objective, samples=samples, raise_at=2, sampler_path=samplers
                ),
                n_trials=4,
            )
        assert len(samples) == 2
        del study

        study, samplers = create_study("resume", False)
        study.optimize(
            functools.partial(objective, samples=samples, sampler_path=samplers),
            n_trials=2,
        )
        assert len(samples) == 4
        assert base_samples == samples

    def test_create_study_starting_from_trial(self, tmp_path: pathlib.Path) -> None:
        @dc.dataclass
        class TestRun:
            sampler_path: pathlib.Path
            samples1: list[float]
            samples2: list[float]
            active_sample: int = 1

            def push_sample(self, s: float) -> None:
                if self.active_sample == 1:
                    self.samples1.append(s)
                else:
                    self.samples2.append(s)

        def objective(trial: optuna.Trial, run: TestRun) -> float:
            hlpo.checkpoint_sampler(trial, run.sampler_path)
            res = trial.suggest_float("accuracy", 0, 1)
            run.push_sample(res)
            hlpo.checkpoint_sampler(trial, run.sampler_path)
            return res

        def get_sampler_names(path: pathlib.Path) -> list[str]:
            s = [p.stem for p in path.glob("*.pkl")]
            s.sort()
            return s

        num_trials = 4
        test_run = TestRun(tmp_path / "resume", [], [])
        test_run.sampler_path.mkdir()

        args: utils.StudyArgs = {}
        args["storage"] = f"sqlite:///{str(tmp_path / 'resume.db')}"
        args["load_if_exists"] = True
        args["study_name"] = "resume"

        study = utils.create_or_load_study(sampler_path=test_run.sampler_path, **args)
        test_run.active_sample = 1
        study.optimize(functools.partial(objective, run=test_run), n_trials=num_trials)
        del study

        base_samplers = get_sampler_names(test_run.sampler_path)
        assert len(base_samplers) == num_trials

        study = utils.create_study_starting_from_trial(
            2, sampler_path=test_run.sampler_path, **args
        )
        bkp_path = tmp_path / "resume_bkp-0"
        assert bkp_path.exists()
        bkp_samplers = get_sampler_names(bkp_path)
        assert bkp_samplers == base_samplers

        pruned_samplers = get_sampler_names(test_run.sampler_path)
        assert len(pruned_samplers) == 2
        assert pruned_samplers == base_samplers[:2]

        test_run.active_sample = 2
        study.optimize(functools.partial(objective, run=test_run), n_trials=2)

        assert len(study.trials) == 4
        assert len(test_run.samples2) == 2
        assert test_run.samples1[2:] == test_run.samples2

        resume_samplers = get_sampler_names(test_run.sampler_path)
        assert resume_samplers == base_samplers



# Getting started tutorial code.
#
# The code is adapted from PyTorch's "Training a Classifier" tutorial:
# https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html

import os
import pathlib
import typing

import torch
import torch.nn.functional as F
import torchvision
import torchvision.transforms.v2 as T
from torch import nn

import helios.core as hlc
import helios.data as hld
import helios.model as hlm
import helios.optim as hlo
import helios.trainer as hlt
from helios.core import logging


class CIFARDataModule(hld.DataModule):
    def __init__(self, root: pathlib.Path) -> None:
        super().__init__()
        self._root = root / "data"

    def prepare_data(self) -> None:
        torchvision.datasets.CIFAR10(root=self._root, train=True, download=True)
        torchvision.datasets.CIFAR10(root=self._root, train=False, download=True)

    def setup(self) -> None:
        # Use the ToImageTensor transform from Helios to automate the conversion from
        # images to tensors.
        transforms = T.Compose(
            [
                hld.transforms.ToImageTensor(),
                T.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        )
        params = hld.DataLoaderParams()
        params.batch_size = 4
        params.shuffle = True
        params.num_workers = 2
        params.drop_last = True
        self._train_dataset = self._create_dataset(
            torchvision.datasets.CIFAR10(
                root=self._root, train=True, download=False, transform=transforms
            ),
            params,
        )

        # The dataloader params are copied when the dataset is created, so we can safely
        # change the options for the validation dataset without interfering with the ones
        # for training.
        params.drop_last = False
        params.shuffle = False
        self._valid_dataset = self._create_dataset(
            torchvision.datasets.CIFAR10(
                root=self._root, train=False, download=False, transform=transforms
            ),
            params,
        )


class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)  # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class ClassifierModel(hlm.Model):
    def __init__(self) -> None:
        super().__init__("classifier")

    def setup(self, fast_init: bool = False) -> None:
        # Note that when we create the network and loss function, we immediately move them
        # to the current device, which has been set by the trainer.
        self._net = Net().to(self.device)
        self._criterion = nn.CrossEntropyLoss().to(self.device)

        # Note that SGD is shipped as part of the default optimizers from Helios, so we
        # can directly request it from create_optimizer instead of building it ourselves.
        self._optimizer = hlo.create_optimizer(
            "SGD", self._net.parameters(), lr=0.001, momentum=0.9
        )

    def load_state_dict(
        self, state_dict: dict[str, typing.Any], fast_init: bool = False
    ) -> None:
        # Note that we don't have to re-map the weights ourselves. They have already been
        # re-mapped for us by the trainer when it loaded the checkpoint.
        self._net.load_state_dict(state_dict["net"])
        self._criterion.load_state_dict(state_dict["criterion"])
        self._optimizer.load_state_dict(state_dict["optimizer"])

    def state_dict(self) -> dict[str, typing.Any]:
        return {
            "net": self._net.state_dict(),
            "criterion": self._criterion.state_dict(),
            "optimizer": self._optimizer.state_dict(),
        }

    def train(self) -> None:
        # If we had more networks, we would shift them to training mode here.
        self._net.train()

    def on_training_start(self) -> None:
        tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

        x = torch.randn((1, 3, 32, 32)).to(self.device)
        tb_logger.add_graph(self._net, x)

    def train_step(self, batch: typing.Any, state: hlt.TrainingState) -> None:
        # Due to the simplicity of the code, we do both the forward and backward passes in
        # the training step, but you could also split it between the train_step and
        # on_training_batch_end if your setup is more complex.
        inputs, labels = batch
        inputs = inputs.to(self.device)
        labels = labels.to(self.device)

        self._optimizer.zero_grad()

        outputs = self._net(inputs)
        loss = self._criterion(outputs, labels)
        loss.backward()
        self._optimizer.step()  # type: ignore[operator]

        # Note that we save the value of the loss function to the _loss_items dictionary.
        # This allows the model to automatically gather the losses for us if we are
        # training in distributed mode.
        self._loss_items["loss"] = loss

    def on_training_batch_end(
        self,
        state: hlt.TrainingState,
        should_log: bool = False,
    ) -> None:
        # If we were training in distributed mode, calling the base Model's function will
        # automatically gather all loss values saved to self._loss_items.
        super().on_training_batch_end(state, should_log)

        # This flag is set to true whenever the number of iterations is a multiple of the
        # logging frequency we set when creating the trainer.
        if should_log:
            root_logger = logging.get_root_logger()
            tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

            loss_val = self._loss_items["loss"]

            root_logger.info(
                f"[{state.global_epoch + 1}, {state.global_iteration:5d}] "
                f"loss: {loss_val:.3f}, "
                f"running loss: {loss_val / state.running_iter:.3f} "
                f"avg time: {state.average_iter_time:.2f}s"
            )
            tb_logger.add_scalar("train/loss", loss_val, state.global_iteration)
            tb_logger.add_scalar(
                "train/running loss",
                loss_val / state.running_iter,
                state.global_iteration,
            )

    def on_training_end(self) -> None:
        # For our example, we're going to save the hyper-params to the tensorboard log so
        # we can compare with other runs.
        # Notice that self._val_scores is active because we validate every epoch. If your
        # validation frequency is different, you may need to alter this code.
        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct // total
        writer = hlc.get_from_optional(logging.get_tensorboard_writer())
        writer.add_hparams(
            {"lr": 0.001, "momentum": 0.9, "epochs": 2},
            {"hparam/accuracy": accuracy, "hparam/loss": self._loss_items["loss"].item()},
        )

    def eval(self) -> None:
        self._net.eval()

    def on_validation_start(self, validation_cycle: int) -> None:
        # The base function will automatically clear the validation scores for us.
        super().on_validation_start(validation_cycle)

        # If you need to add further data to the table, you can do so here. In our case,
        # we're going to add the total number of labels seen and how many of those were
        # correct so we can compute the accuracy metric.
        self._val_scores["total"] = 0
        self._val_scores["correct"] = 0

    def valid_step(self, batch: typing.Any, step: int) -> None:
        images, labels = batch
        images = images.to(self.device)
        labels = labels.to(self.device)

        outputs = self._net(images)

        _, predicted = torch.max(outputs.data, 1)
        self._val_scores["total"] += labels.size(0)
        self._val_scores["correct"] += (predicted == labels).sum().item()

    def on_validation_end(self, validation_cycle: int) -> None:
        root_logger = logging.get_root_logger()
        tb_logger = hlc.get_from_optional(logging.get_tensorboard_writer())

        # Grab the validation scores and compute the accuracy metric so we can log it. If
        # we were in distributed mode, you would need to gather the values here.
        total = self._val_scores["total"]
        correct = self._val_scores["correct"]
        accuracy = 100 * correct // total

        root_logger.info(f"[Validation {validation_cycle}] accuracy: {accuracy}")
        tb_logger.add_scalar("val", accuracy, validation_cycle)


if __name__ == "__main__":
    # Set the CUBLAS workspace setting to allow determinism to be used in CUDA >= 10.2.
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

    datamodule = CIFARDataModule(pathlib.Path.cwd())
    model = ClassifierModel()

    trainer = hlt.Trainer(
        run_name="cifar10",
        train_unit=hlt.TrainingUnit.EPOCH,
        total_steps=2,
        valid_frequency=1,
        chkpt_frequency=1,
        print_frequency=10,
        enable_tensorboard=True,
        enable_file_logging=True,
        enable_progress_bar=True,
        enable_deterministic=True,
        chkpt_root=pathlib.Path.cwd() / "chkpt",
        log_path=pathlib.Path.cwd() / "logs",
        run_path=pathlib.Path.cwd() / "runs",
    )

    trainer.fit(model, datamodule)



"""
Loss package for Helios.

This contains the classes, utilities, and registries related to loss functions.
"""

from .utils import LOSS_REGISTRY, create_loss
from .weighted_loss import WeightedLoss

__all__ = ["LOSS_REGISTRY", "create_loss", "WeightedLoss"]



import collections as col
import math

from torch import optim
from torch.optim import lr_scheduler

from .utils import SCHEDULER_REGISTRY


def _get_position_from_periods(iteration: int, cummulative_period: list[int]) -> int:
    """
    Get position from a period list.

    Specifically, it returns the index of the right-closest number in the period list.
    For example, suppose ``cummulative_period`` is ``[100, 200, 300, 400]``. Then:
    * If ``iteration == 50``, return 0
    * If ``iteration == 210``, return 2
    * If ``iteration == 300``, return 2.

    Args:
        iteration: current iteration.
        cummulative_period: cummulative period list.

    Returns:
        The position of the right-closest number in the period list
    """
    for i, period in enumerate(cummulative_period):
        if iteration <= period:
            return i

    return 0


@SCHEDULER_REGISTRY.register
class CosineAnnealingRestartLR(lr_scheduler.LRScheduler):
    """
    A cosine annealing with restarts LR scheduler.

    Example:
        Given

        .. code-block:: text

            periods = [10, 10, 10, 10]
            restart_weights = [1, 0.5, 0.5, 0.5]
            eta_min = 1e-7

        Then the scheduler will have 4 cycles of 10 iterations each. At the 10th, 20th,
        and 30th, the scheduler will restart with the weights in ``restart_weights``.

    Args:
        optimizer: the optimizer.
        periods: period for each cosine annealing cycle.
        restart_weights: (optional) restarts weights at each restart iteration.
        eta_min: The minimum lr. Defaults to 0
        last_epoch: Used in _LRScheduler. Defaults to -1.
    """

    def __init__(
        self,
        optimizer: optim.Optimizer,
        periods: list[int],
        restart_weights: list[int] | None = None,
        eta_min: float = 0,
        last_epoch: int = -1,
    ):
        """Create the scheduler."""
        if restart_weights is None:
            restart_weights = [1]

        self._periods = periods
        self._restart_weights = restart_weights
        self._eta_min = eta_min
        assert len(self._periods) == len(
            self._restart_weights
        ), "periods and restart_weights should have the same length."
        self._cumulative_period = [
            sum(self._periods[0 : i + 1]) for i in range(len(self._periods))
        ]
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        """Return the current learning rate."""
        idx = _get_position_from_periods(self.last_epoch, self._cumulative_period)
        current_weight = self._restart_weights[idx]
        nearest_restart = 0 if idx == 0 else self._cumulative_period[idx - 1]
        current_period = self._periods[idx]

        return [
            self._eta_min
            + current_weight
            * 0.5
            * (base_lr - self._eta_min)
            * (
                1
                + math.cos(
                    math.pi * ((self.last_epoch - nearest_restart) / current_period)
                )
            )
            for base_lr in self.base_lrs
        ]


@SCHEDULER_REGISTRY.register
class MultiStepRestartLR(lr_scheduler.LRScheduler):
    """
    Multi-step with restarts LR scheduler.

    Args:
        optimizer: torch optimizer.
        milestones: iterations that will decrease learning rate.
        gamma: decrease ratio. Defaults to 0.1.
        restarts: (optional) restart iterations.
        restart_weights: (optional) restart weights at each restart iteration.
        last_epoch: used in _LRScheduler. Defaults to -1.
    """

    def __init__(
        self,
        optimizer: optim.Optimizer,
        milestones: list[int],
        gamma: float = 0.1,
        restarts: list[int] | None = None,
        restart_weights: list[int] | None = None,
        last_epoch: int = -1,
    ):
        """Create the scheduler."""
        if restarts is None:
            restarts = [0]
        if restart_weights is None:
            restart_weights = [1]

        self._milestones = col.Counter(milestones)
        self._gamma = gamma
        self._restarts = restarts
        self._restart_weights = restart_weights
        assert len(self._restarts) == len(
            self._restart_weights
        ), "restarts and their weights do not match."
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        """Return the current learning rate."""
        if self.last_epoch in self._restarts:
            weight = self._restart_weights[self._restarts.index(self.last_epoch)]
            return [group["initial_lr"] * weight for group in self.optimizer.param_groups]
        if self.last_epoch not in self._milestones:
            return [group["lr"] for group in self.optimizer.param_groups]
        return [
            group["lr"] * self._gamma ** self._milestones[self.last_epoch]
            for group in self.optimizer.param_groups
        ]



import pathlib

import pytest
import torch
import torch.nn.functional as F
from torch import nn

from helios import nn as hln
from helios import onnx
from helios.nn import layers as hlnl


@hln.NETWORK_REGISTRY.register
class ClassifierNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)  # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class TestNewtorks:
    def test_registry(self, check_registry) -> None:
        check_registry(hln.NETWORK_REGISTRY, ["ClassifierNet"])

    def test_create(self, check_create_function) -> None:
        check_create_function(hln.NETWORK_REGISTRY, hln.create_network)

    def test_ema(self) -> None:
        net = nn.Conv2d(1, 20, 5)
        ema = hln.swa_utils.EMA(net)
        ema.update(net)
        n = ema.module
        assert net != n


class TestAdaptiveAvgPool2D:
    def check_input_exception(self, *args) -> None:
        with pytest.raises(RuntimeError):
            hlnl.AdaptiveAvgPool2d(*args)

    def test_invalid_inputs(self) -> None:
        self.check_input_exception((1, 2, 3))
        self.check_input_exception((1,))
        self.check_input_exception(1.0)
        self.check_input_exception((1, 1.0))
        self.check_input_exception((1.0, 1))
        self.check_input_exception(None)

    def test_shape_tuple(self) -> None:
        m_t = nn.AdaptiveAvgPool2d((5, 7))
        m_h = hlnl.AdaptiveAvgPool2d((5, 7))
        x = torch.randn(1, 64, 8, 9)

        base = m_t(x)
        ret = m_h(x)
        assert base.shape == ret.shape

    def test_shape_int(self) -> None:
        m_t = nn.AdaptiveAvgPool2d(7)
        m_h = hlnl.AdaptiveAvgPool2d(7)
        x = torch.randn(1, 64, 10, 9)

        base = m_t(x)
        ret = m_h(x)
        assert base.shape == ret.shape

    def test_none_tuple_x_axis(self) -> None:
        m_t = nn.AdaptiveAvgPool2d((None, 7))
        m_h = hlnl.AdaptiveAvgPool2d((None, 7))
        x = torch.randn(1, 64, 10, 9)

        base = m_t(x)
        ret = m_h(x)
        assert base.shape == ret.shape

    def test_none_tuple_y_axis(self) -> None:
        m_t = nn.AdaptiveAvgPool2d((7, None))
        m_h = hlnl.AdaptiveAvgPool2d((7, None))
        x = torch.randn(1, 64, 10, 9)

        base = m_t(x)
        ret = m_h(x)
        assert base.shape == ret.shape

    def check_export(
        self, model: nn.Module, x: torch.Tensor, out_path: pathlib.Path
    ) -> None:
        onnx.export_to_onnx(model, x, out_path, validate_output=True)
        assert out_path.exists()

    @pytest.mark.filterwarnings("ignore:Converting")
    def test_export(self, tmp_path: pathlib.Path) -> None:
        self.check_export(
            hlnl.AdaptiveAvgPool2d((5, 7)),
            torch.randn(1, 64, 8, 9),
            tmp_path / "5_7.onnx",
        )
        self.check_export(
            hlnl.AdaptiveAvgPool2d(7),
            torch.randn(1, 64, 10, 9),
            tmp_path / "7_square.onnx",
        )
        self.check_export(
            hlnl.AdaptiveAvgPool2d((None, 7)),
            torch.randn(1, 64, 10, 9),
            tmp_path / "none_x.onnx",
        )

        self.check_export(
            hlnl.AdaptiveAvgPool2d((7, None)),
            torch.randn(1, 64, 10, 9),
            tmp_path / "none_y.onnx",
        )



import enum
import gc
import pathlib
import pickle
import shutil
import typing
import warnings

import optuna


class _BackupType(enum.Enum):
    FILE = enum.auto()
    DIR = enum.auto()


class StudyArgs(typing.TypedDict):
    """
    Type-hint class for the arguments of ``optuna.create_study``.

    See the `docs
    <https://optuna.readthedocs.io/en/latest/reference/generated/optuna.create_study.html>`__
    for more information.

    Args:
        storage: Database URL. If this argument is set to None, ``InMemoryStorage`` is
            used, and the ``Study`` will not be persistent.
        sampler: A sampler object that implements background algorithm for value
            suggestion.
        pruner: A pruner object that decides early stopping for unpromising trials.
        study_name: Study's name.
        direction: Direction of optimization.
        load_if_exists: flag to control the behaviour to handle conflict of study names.
        directions: A sequence of directions during multi-objective optimization.
    """

    storage: typing.NotRequired[str | optuna.storages.BaseStorage | None]
    sampler: typing.NotRequired[optuna.samplers.BaseSampler | None]
    pruner: typing.NotRequired[optuna.pruners.BasePruner | None]
    study_name: typing.NotRequired[str | None]
    direction: typing.NotRequired[str | optuna.study.StudyDirection]
    load_if_exists: typing.NotRequired[bool]
    directions: typing.NotRequired[
        typing.Sequence[str | optuna.study.StudyDirection] | None
    ]


def _get_backup_name(root: pathlib.Path, name: str, bkp_type: _BackupType) -> str:
    i = -1
    for path in root.iterdir():
        if not path.is_file() and bkp_type == _BackupType.FILE:
            continue
        if not path.is_dir() and bkp_type == _BackupType.DIR:
            continue
        if "_bkp-" in path.stem:
            elems = path.stem.split("_")
            idx = [i for i, e in enumerate(elems) if "bkp-" in e][0]
            n = int(elems[idx].split("-")[-1])
            i = max(i, n)

    i = i + 1
    return f"{name}_bkp-{i}"


def _backup_study(storage_path: pathlib.Path) -> None:
    if not storage_path.exists():
        return

    name = storage_path.stem
    root = storage_path.parent
    bkp_name = _get_backup_name(root, name, _BackupType.FILE)
    bkp_path = root / f"{bkp_name}.db"
    shutil.copy(str(storage_path), str(bkp_path))


def _backup_sampler_chkpts(sampler_path: pathlib.Path) -> None:
    if not sampler_path.exists():
        return

    name = sampler_path.stem
    root = sampler_path.parent
    bkp_name = _get_backup_name(root, name, _BackupType.DIR)
    bkp_path = root / bkp_name
    bkp_path.mkdir()

    for chkpt in sampler_path.glob("*.pkl"):
        dst_path = bkp_path / chkpt.name
        shutil.copy2(str(chkpt), str(dst_path))


def _sampler_chkpt_key(path: pathlib.Path) -> int:
    return int(path.stem.split("-")[-1])


def _check_study_args(**study_args: typing.Unpack[StudyArgs]) -> None:
    if "storage" not in study_args or study_args["storage"] is None:
        raise RuntimeError("error: RDB storage is required for resuming studies")
    if "load_if_exists" not in study_args or not study_args["load_if_exists"]:
        raise RuntimeError(
            "error: study must be created with 'load_if_exists' set to true"
        )
    if not isinstance(study_args["storage"], str):
        raise TypeError("error: only strings are supported for 'storage'")

    if "sqlite:///" not in study_args["storage"]:
        raise RuntimeError("error: only sqlite storage is supported")


def _get_storage_path(**study_args: typing.Unpack[StudyArgs]) -> pathlib.Path:
    name = study_args["storage"]
    assert isinstance(name, str)
    return pathlib.Path(name.removeprefix("sqlite:///")).resolve()


def checkpoint_sampler(trial: optuna.Trial, chkpt_root: pathlib.Path) -> None:
    """
    Create a checkpoint with the state of the sampler.

    This function can be used to ensure that if a study is restarted, the state of the
    sampler is recovered so trials can be reproducible. The function will automatically
    create a checkpoint using ``torch.save``.

    .. note::
        It is recommended that this function be called at the start and end of the
        objective function to ensure the state of the sampler is correctly stored. This
        ensures that if a trial fails before it returns, then the sampler can be restored
        and the same values can be retrieved. Alternatively, if the trial completes and
        you wish to resume the study, then the sampler can be restored to the state it
        would've had if the study hadn't been stopped.

    Args:
        trial: the current trial.
        chkpt_root: the root where the checkpoints will be saved.
    """
    chkpt_path = chkpt_root / (f"sampler_trial-{trial.number}.pkl")

    sampler = trial.study.sampler
    with chkpt_path.open("wb+") as outfile:
        pickle.dump(sampler, outfile)


def restore_sampler(
    chkpt_root: pathlib.Path, trial_id: int | None = None
) -> optuna.samplers.BaseSampler | None:
    """
    Restore the sampler from a previously saved checkpoint.

    This function can be used in tandem with
    :py:func:`~helios.plugins.optuna.checkpoint_sampler` to ensure that the last
    checkpoint is loaded and the correct state is restored for the sampler. This function
    **needs** to be called before ``optuna.create_study`` is called.

    If you wish to restore the sampler for a specific trial, you can use ``trial_id`` to
    restore it.

    Args:
        chkpt_root: the root where the checkpoints are stored.
        trial_id: the trial ID for which to restore the sampler.

    Returns:
        The restored sampler.
    """
    chkpts = list(chkpt_root.glob("*.pkl"))
    chkpts.sort(key=_sampler_chkpt_key)
    if len(chkpts) == 0:
        return None

    chkpt = (
        chkpts[trial_id]
        if trial_id is not None and trial_id < len(chkpts)
        else chkpts[-1]
    )

    sampler = pickle.load(chkpt.open("rb"))
    return sampler


def create_or_load_study(
    sampler_path: pathlib.Path | None = None,
    backup_study: bool = False,
    backup_samplers: bool = False,
    **study_args: typing.Unpack[StudyArgs],
) -> optuna.Study:
    """
    Start or continue a study.

    This function can be used to either start a new study or resume one that has stopped.
    This leverages the existing built-in checkpoint system from Helios as well as
    :py:func:`~helios.plugins.optuna.restore_sampler` to ensure the study can be resumed
    properly.

    .. warning::
        This function **requires** the following conditions to be true:
            #. The study uses a string to specify the ``RDBStorage`` as the argument to
                ``storage`` for ``optuna.create_study``.
            #. The storage string starts with ``sqlite:///``.
            #. ``load_if_exists`` is set to true.

    If you have been using :py:func:`~helios.plugins.optuna.checkpoint_sampler` and wish
    to restore the sampler upon restarting the study, you can provide the root where the
    checkpoints are stored by passing in ``sampler_path``.

    You can use ``backup_study`` to create a backup of the original study. If there is
    one, the backup will be called ``<study-name>_backup-#`` where ``<study-name>`` is the
    name of the database of the original study and ``#`` is an incremental number starting
    at 0.

    .. note::
        This function **does not** retry failed trials, nor does it continue trials that
        were stopped. It will simply continue with the next (new) trial as determined by
        Optuna. To restart a study from a specific trial, please use
        :py:func:`~helios.plugins.optuna.create_study_starting_from_trial`.

    Args:
        sampler_path: the path where the sampler checkpoints are stored. Defaults to None.
        backup_study: if True, the original study is backed up so it can be re-used later
            on. Defaults to false.
        backup_samplers: if True, a backup will be made of the samplers checkpoint folder
            (if given). Defaults to false.
        study_args: arguments for ``optuna.create_study``.
    """
    _check_study_args(**study_args)

    storage = _get_storage_path(**study_args)
    if backup_study:
        _backup_study(storage)

    if backup_samplers and sampler_path is not None:
        _backup_sampler_chkpts(sampler_path)

    sampler = restore_sampler(sampler_path) if sampler_path is not None else None
    if "sampler" in study_args and sampler is not None:
        warnings.warn(
            "warning: existing sampler will be replaced with checkpointed sampler",
            stacklevel=2,
        )

    # Only set the sampler if we actually have one, otherwise we'll end up overriding the
    # user's sampler and that's not what we want.
    if sampler is not None:
        study_args["sampler"] = sampler
    study = optuna.create_study(**study_args)
    return study


def create_study_starting_from_trial(
    trial_id: int,
    backup_study: bool = True,
    backup_samplers: bool = True,
    sampler_path: pathlib.Path | None = None,
    **study_args: typing.Unpack[StudyArgs],
) -> optuna.Study:
    """
    Resume a study from a specific trial number.

    The goal of this function is to allow studies to resume from a specific trial number,
    effectively discarding all trials after that point. This can be used to recover from
    certain types of unexpected errors that result in trials being invalid but still
    marked as completed, thereby making :py:func:`~helios.plugins.optuna.resume_study`
    unable to properly resume. To accomplish this, the function will do the following:

    #. Grab all the trials from the study created by the given ``study_args`` and grab all
        trials whose numbers are *less than or equal to* ``last_trial_number``.
    #. Create a new study with the same name and storage. This new study will get all of
        the trials that were obtained in the previous step.

    If ``sampler_path`` is provided, then the samplers for trials that occur after
    ``trial_id`` will be pruned. The sampler that corresponds to ``trial_id`` will be
    preserved so it can be restored in to the new study. If you don't wish this to happen,
    simply pass in the sampler in ``study_args`` and set ``sampler_path`` to None.

    .. note::
        Trials are 0-indexed. This means that trial ``N`` maps to ID ``N - 1``.

    .. warning::
        This function **requires** the following conditions to be true:

        #. It is called **before** the trials are started.
        #. The study uses ``RDBStorage`` in ``study_args``.
        #. ``load_if_exists`` is set to True in ``study_args``.

    By default, the original study and samplers checkpoint folder will be backed up
    (assuming they exist). The backups are named ``<name>_backup-#`` where ``<name>`` is
    either the name of the database of the original study or the sampler folder, and ``#``
    is an incremental number starting at 0. This behaviour can be disabled by setting
    ``backup_study`` or ``backup_samplers`` to false.

    In addition, this function can also remove the saved samplers from the discarded
    trials if ``sampler_root`` is provided.

    Args:
        trial_id: the ID of the last trial from which the study should resume.
        backup_study: if True, the original study is backed up so it can be re-used later
            on.
        sampler_path: the path where the sampler checkpoints are stored.
        study_args: arguments for ``optuna.create_study``.

    Returns:
        The study with trials up to ``last_trial_number``.
    """
    _check_study_args(**study_args)

    storage = _get_storage_path(**study_args)
    if backup_study:
        _backup_study(storage)

    if backup_samplers and sampler_path is not None:
        _backup_sampler_chkpts(sampler_path)

    study = optuna.create_study(**study_args)

    # Early exit: if there are no trials, return immediately.
    if len(study.trials) == 0:
        return study

    valid_trials: list[optuna.trial.FrozenTrial] = []
    for trial in study.trials:
        if trial.number < trial_id:
            valid_trials.append(trial)

    if sampler_path is not None:
        chkpts = list(sampler_path.glob("*.pkl"))
        chkpts.sort(key=_sampler_chkpt_key)
        for i, path in enumerate(chkpts):
            if i < trial_id:
                continue
            path.unlink()

        if "sampler" in study_args:
            warnings.warn(
                "warning: existing sampler will be replaced with checkpointed sampler",
                stacklevel=2,
            )

        study_args["sampler"] = restore_sampler(sampler_path, trial_id)

    del study
    gc.collect()

    tmp_storage = storage.parent / (storage.stem + "_tmp" + storage.suffix)
    storage.rename(tmp_storage)
    study = optuna.create_study(**study_args)

    for trial in valid_trials:
        study.add_trial(trial)

    tmp_storage.unlink()
    return study



"""
Data package for Helios.

Holds all functions and classes related to datasets, dataloaders, and data processing.
"""

from .datamodule import (
    COLLATE_FN_REGISTRY,
    DATASET_REGISTRY,
    DataLoaderParams,
    DataModule,
    DatasetSplit,
    create_collate_fn,
    create_dataloader,
    create_dataset,
)
from .transforms import TRANSFORM_REGISTRY, create_transform

__all__ = [
    "COLLATE_FN_REGISTRY",
    "DATASET_REGISTRY",
    "DataLoaderParams",
    "DataModule",
    "DatasetSplit",
    "create_collate_fn",
    "create_dataloader",
    "create_dataset",
    "TRANSFORM_REGISTRY",
    "create_transform",
]



import pathlib

import torch

import helios
import helios.core as hlc
import helios.trainer as hlt
from helios.chkpt_migrator import migrate_checkpoints_to_current_version


class TestChkptMigrator:
    def test_from_pre_release(self, tmp_path: pathlib.Path) -> None:
        state = {
            "training_state": hlt.TrainingState().dict(),
            "model": {},
            "rng": {},
        }

        torch.save(state, tmp_path / "state.pth")
        migrate_checkpoints_to_current_version(tmp_path)

        new_state = hlc.safe_torch_load(tmp_path / "state.pth")
        assert all(
            key in new_state for key in ("version", "training_state", "model", "rng")
        )
        assert new_state["version"] == helios.__version__

    def test_from_pre_1_1(self, tmp_path: pathlib.Path) -> None:
        state = {
            "training_state": hlt.TrainingState().dict(),
            "model": {},
            "rng": {},
            "version": helios.__version__,
        }

        torch.save(state, tmp_path / "state.pth")
        migrate_checkpoints_to_current_version(tmp_path)

        new_state = hlc.safe_torch_load(tmp_path / "state.pth")
        assert isinstance(new_state["training_state"], hlt.TrainingState)

    def test_from_current_release(self, tmp_path: pathlib.Path) -> None:
        state = {
            "training_state": hlt.TrainingState(),
            "model": {},
            "rng": {},
            "version": helios.__version__,
        }

        torch.save(state, tmp_path / "state.pth")
        migrate_checkpoints_to_current_version(tmp_path)

        new_state = hlc.safe_torch_load(tmp_path / "state.pth")
        assert state == new_state



import torch
from torch import cuda


def requires_cuda_support() -> None:
    """
    Ensure that CUDA support is found, or raise an exception otherwise.

    Raises:
        RuntimeError: if no CUDA support is found.

    """
    if not cuda.is_available():
        raise RuntimeError("error: expected CUDA support but none was found")


class DisableCuDNNBenchmarkContext:
    """
    Allow disabling CuDNN benchmark within a scope.

    The intention is to facilitate the disabling of CuDNN benchmark for specific purposes
    (such as validation or testing) but then restoring it to its previous state upon
    leaving the scope. Note that if CUDA is not available, the scope does nothing.

    Example:
        .. code-block:: python

            torch.backends.cudnn.benchmark = True # Enable CuDNN
            ...
            with DisableCuDNNBenchmarkContext():
                # Benchmark is disabled.
                print(torch.backends.cudnn.benchmark) # <- Prints False
                ...

            print(torch.backends.cudnn.benchmark) # <- Prints true
    """

    def __init__(self) -> None:
        """Create the CuDNN context."""
        if cuda.is_available():
            self._init_value = torch.backends.cudnn.benchmark

    def __enter__(self) -> None:
        """Disable CuDNN benchmark."""
        if cuda.is_available():
            torch.backends.cudnn.benchmark = False

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        """Restore CuDNN benchmark to its starting state."""
        if cuda.is_available():
            torch.backends.cudnn.benchmark = self._init_value



import pathlib

import cv2
import numpy as np
import numpy.typing as npt
import PIL
import torch


def load_image(
    path: pathlib.Path, flags: int = cv2.IMREAD_COLOR, as_rgb: bool = True
) -> npt.NDArray:
    """
    Load the given image using OpenCV.

    ``flags`` correspond to the ``cv2.IMREAD_`` flags from OpenCV. Please see the full
    list of options
    `here <https://docs.opencv.org/3.4/d8/d6a/group__imgcodecs__flags.html>`__. If no
    value is passed, the image will be loaded using ``cv2.IMREAD_COLOR``, which will load
    it as an 8-bit BGR image.

    ``as_rgb`` can be used to automatically convert the image from OpenCV's default BGR to
    RGB using the following logic:
    * If the image is grayscale, then it is returned as is.
    * If the image is a 3-channel BGR, it is converted to RGB.
    * If the image is a 4-channel BGRA, it is converted to RGBA.
    If all these checks fail, the image is returned as is.

    Args:
        path: the path to the image to load.
        flags: the OpenCV flags to use when loading.
        as_rgb: if true, the image will be converted from BGR/BGRA to RGB/RGBA, otherwise
            the image is returned as is.

    Returns:
        The loaded image.
    """
    img = cv2.imread(str(path), flags=flags)
    if not as_rgb:
        return img

    if len(img.shape) == 2:
        return img

    c = img.shape[2]
    if c == 3:
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    if c == 4:
        return cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)

    raise RuntimeError(
        f"error: expected a 3 or 4 channel image but received {c} channels"
    )


def load_image_pil(
    path: pathlib.Path, out_fmt: str = "", as_numpy: bool = True
) -> npt.NDArray | PIL.Image.Image:
    """
    Load the given image using PIL.

    ``out_fmt`` is a format string that can be passed in to PIL.Image.convert. Please
    `here <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.convert>`__
    for the list of accepted strings.
    If no string is passed, the image will be converted to RGB format.
    By default, the output is a NumPY array. If you need a PIL image instead, set
    ``as_numpy`` to false.

    Args:
        path: the path to the image to load.
        out_fmt: the format to convert the loaded image to. Defaults to empty.
        as_numpy: if true, the loaded image will be returned as a NumPY array, otherwise
            it is returned as a PIL image. Defaults to true.

    Returns:
        The loaded image.
    """
    with path.open(mode="rb") as infile:
        img = PIL.Image.open(infile)
        out = img.convert(out_fmt) if out_fmt != "" else img.convert("RGB")
        if as_numpy:
            return np.array(out)
        return out


def tensor_to_numpy(
    x: torch.Tensor,
    squeeze: bool = True,
    clamp: tuple[float, float] | None = (0, 1),
    transpose: bool = True,
    dtype: torch.dtype | None = torch.float,
    as_uint8: bool = True,
) -> npt.NDArray:
    r"""
    Convert the given tensor to a numpy array.

    Args:
        x: the tensor to convert.
        squeeze: if true, squeeze to remove all dimensions of size 1. Defaults to true.
        clamp: tuple containing min/max values to clamp the tensor to. Can be ``None`` if
            no clamping is desired. Defaults to ``(0, 1)``.
        transpose: if true, transpose the tensor so the dimensions are :math:`H \times W
            \times C` or :math:`B \times H \times W \times C`. Defaults to true.
        dtype: the type to convert the tensor to. Can be set to ``None`` if no conversion
            is desired. Defaults to ``torch.float``.
        as_uint8: if true, convert the final array to be of type uint8 and in the range
            :math:`[0, 255]`. Defaults to true.

    Returns:
        The converted array.
    """
    x = x.clone()
    if squeeze:
        x.squeeze_()
    if clamp is not None:
        x.clamp_(min=clamp[0], max=clamp[1])
    if dtype is not None:
        x = x.to(dtype)
    as_np = x.cpu().detach().numpy()

    if transpose:
        as_np = (
            np.transpose(as_np, (1, 2, 0))
            if as_np.ndim == 3
            else np.transpose(as_np, (0, 2, 3, 1))
            if as_np.ndim == 4
            else as_np
        )

    if as_uint8:
        as_np = np.uint8((as_np * 255.0).round())  # type: ignore[assignment]
    return as_np


def show_tensor(tens: torch.Tensor, title: str = "debug window") -> None:
    """
    Display the image held by the tensor. Useful for debugging purposes.

    Args:
        tens: the image tensor to display in range :math:`[0, 1]`.
        title: the title of the displayed window. Defaults to "debug window".
    """
    img = tensor_to_numpy(tens)
    img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    cv2.imshow(title, img)
    cv2.waitKey(0)


def show_tensors(tens: torch.Tensor) -> None:
    """
    Show batches of tensors.

    Args:
        tens: the batch of tensors to display in range :math:`[0, 1]`.
    """
    if len(tens.shape) == 3:
        show_tensor(tens)
        return

    for b in range(tens.shape[0]):
        show_tensor(tens[b], title=f"tensor[{b}]")


def convert_to_hwc(img: npt.NDArray, input_order: str = "HWC") -> npt.NDArray:
    """
    Change the order of the input image channels so the result is in (h, w, c) order.

    If the input image is a single-channel image, then the return is (h, w, 1).

    Args:
        img: input image.
        input_order: the order of the channels of the input image. Must be one of 'HWC'
            or 'CHW'.

    Returns:
        The shuffled image.
    """
    assert input_order in ("HWC", "CHW")

    if len(img.shape) == 2:
        img = img[..., None]
    if input_order == "CHW":
        img = img.transpose(1, 2, 0)
    return img


def to_y_channel(img: npt.NDArray) -> npt.NDArray:
    """
    Return the Y (luma) channel of a YCbCr image.

    Args:
        img: input image in YCbCr format. Must be in the range :math:`[0, 255]`.

    Returns:
        The luma channel of the input image.
    """
    img = img.astype(np.float32) / 255.0
    if img.ndim == 3 and img.shape[2] == 3:
        img = bgr2ycbcr(img, only_y=True)
        img = img[..., None]
    return img * 255.0


def bgr2ycbcr(img: npt.NDArray, only_y: bool = False) -> npt.NDArray:
    """
    Convert the given numpy image array from BGR to YCBCR.

    If only the Y channel is required, set ``only_y`` to true.

    Args:
        img: the BGR image to convert.
        only_y: if true, only the luma (Y) channel will be returned.

    Returns:
        The converted image.
    """
    intype = img.dtype
    img.astype(np.float32)
    if intype != np.uint8:
        img *= 255

    if only_y:
        rlt = np.dot(img, [24.966, 128.553, 65.481]) / 255.0 + 16.0
    else:
        rlt = np.matmul(
            img,
            [
                [24.966, 112.0, -18.214],
                [128.553, -74.203, -93.786],
                [65.481, -37.797, 112.0],
            ],
        ) / 255.0 + [16, 128, 128]

    if intype == np.uint8:
        rlt = rlt.round()
    else:
        rlt /= 255.0
    return rlt.astype(intype)


def rgb2ycbcr_torch(img: torch.Tensor, only_y: bool = False) -> torch.Tensor:
    """
    Convert the given torch Tensor image array from RGB to YCBCR.

    If only the Y channel is required, set ``only_y`` to true.

    Args:
        img: the BGR image to convert.
        only_y: if true, only the luma (Y) channel will be returned.

    Returns:
        The converted image.
    """
    if only_y:
        weight = torch.tensor([[65.481], [128.553], [24.966]]).to(img)
        out_img = torch.matmul(img.permute(0, 2, 3, 1), weight).permute(0, 3, 1, 2) + 16.0
    else:
        weight = torch.tensor(
            [
                [65.481, -37.797, 112.0],
                [128.553, -74.203, -93.786],
                [24.966, 112.0, -18.214],
            ]
        ).to(img)
        bias = torch.tensor([16, 128, 128]).view(1, 3, 1, 1).to(img)
        out_img = torch.matmul(img.permute(0, 2, 3, 1), weight).permute(0, 3, 1, 2) + bias

    out_img = out_img / 255.0
    return out_img



import typing

import optuna
import torch

import helios.core.distributed as dist
import helios.model as hlm
import helios.plugins as hlp
import helios.trainer as hlt

_PRUNED_KEY = "ddp_hl:pruned"
_CYCLE_KEY = "ddp_hl:cycle"

# Ignore private member access
# ruff: noqa: SLF001


@hlp.PLUGIN_REGISTRY.register
class OptunaPlugin(hlp.Plugin):
    """
    Plug-in to do hyper-parameter tuning with Optuna.

    This plug-in integrates `Optuna <https://optuna.readthedocs.io/en/stable/>`__ into the
    training system in order to provide hyper-parameter tuning. The plug-in provides the
    following functionality:

    #. Automatic handling of trial pruning.
    #. Automatic reporting of metrics.
    #. Exception registration for trial pruning.
    #. Easy integration with Helios' checkpoint system to continue stopped trials.

    Example:
        .. code-block:: python

            import helios.plugins as hlp
            import optuna

            def objective(trial: optuna.Trial) -> float:
                datamodule = ...
                model = ...
                plugin = hlp.optuna.OptunaPlugin(trial, "accuracy")

                trainer = ...

                # Automatically registers the plug-in with the trainer.
                plugin.configure_trainer(trainer)

                # This can be skipped if you don't want the auto-resume functionality or
                # if you wish to manage it yourself.
                plugin.configure_model(model)

                trainer.fit(model, datamodule)
                plugin.check_pruned()
                return model.metrics["accuracy"]

            def main():
                # Note that the plug-in requires the storage to be persistent.
                study = optuna.create_study(storage="sqlite:///example.db", ...)
                study.optimize(objective, ...)

    Args:
        trial: the Optuna trial.
        metric_name: the name of the metric to monitor. This assumes the name will be
            present in the :py:attr:`~helios.model.model.Model.metrics` table.
    """

    plugin_id = "optuna"

    def __init__(self, trial: optuna.Trial, metric_name: str) -> None:
        """Create the plug-in."""
        super().__init__(self.plugin_id)
        self._trial = trial
        self._metric_name = metric_name
        self._last_cycle: int = 0

        self.unique_overrides.should_training_stop = True

    @property
    def trial(self) -> optuna.Trial:
        """Return the trial."""
        return self._trial

    @trial.setter
    def trial(self, t: optuna.Trial) -> None:
        self._trial = t

    def configure_trainer(self, trainer: hlt.Trainer) -> None:
        """
        Configure the trainer with the required settings.

        This will do two things:

        #. Register the plug-in itself with the trainer.
        #. Append the trial pruned exception to the trainer.

        Args:
            trainer: the trainer instance.
        """
        self._register_in_trainer(trainer)
        self._append_train_exceptions(optuna.TrialPruned, trainer)

    def configure_model(self, model: hlm.Model) -> None:
        """
        Configure the model to set the trial number into the save name.

        This will alter the :py:attr:`~helios.model.model.Model.save_name` property of the
        model by appending :code:`_trial-<trial-numer>`.

        Args:
            model: the model instance.
        """
        n_trial = self.trial.number
        model._save_name = model._save_name + f"_trial-{n_trial}"

    def suggest(self, type_name: str, name: str, **kwargs: typing.Any) -> typing.Any:
        """
        Generically Wrap the ``suggest_`` family of functions of the optuna trial.

        This function can be used to easily invoke the corresponding ``suggest_`` function
        from the Optuna trial held by the plug-in without having to manually type each
        individual function. This lets you write generic code that can be controlled by an
        external source (such as command line arguments or a config table). The function
        wraps the following functions:

        .. list-table:: Suggestion Functions
            :header-rows: 1

            * - Function
              - Name
            * - ``optuna.Trial.suggest_categorical``
              - categorical
            * - ``optuna.Trial.suggest_int``
              - int
            * - ``optuna.Trial.suggest_float``
              - float

        .. warning::
            Functions that are marked as deprecated by Optuna are *not* included in this
            wrapper.

        .. note::
            You can find the exact arguments for each function `here
            <https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html>`__.

        Example:
            .. code-block:: python

                import helios.plugin as hlp
                import optuna

                def objective(trial: optuna.Trial) -> float:
                    plugin = hlp.optuna.OptunaPlugin(trial, "accuracy")
                    # ... configure model and trainer.

                    val1 = plugin.suggest("categorical", "val1", choices=[1, 2, 3])
                    val2 = plugin.suggest("int", "val2", low=0, high=10)
                    val3 = plugin.suggest("float", "val3", low=0, high=1)

        Args:
            type_name: the name of the type to suggest from.
            name: a parameter name
            **kwargs: keyword arguments to the corresponding suggest function.

        Raises:
            KeyError: if the value passed in to ``type_name`` is not recognised.
        """
        if type_name not in ("categorical", "float", "int"):
            raise KeyError(f"error: {type_name} is not a valid suggestion type.")

        fn = getattr(self._trial, f"suggest_{type_name}")
        return fn(name, **kwargs)

    def setup(self) -> None:
        """
        Configure the plug-in.

        Raises:
            ValueError: if the study wasn't created with persistent storage.
        """
        if self.is_distributed and not (
            isinstance(self.trial.study._storage, optuna.storages._CachedStorage)
            and isinstance(self.trial.study._storage._backend, optuna.storages.RDBStorage)
        ):
            raise ValueError(
                "error: optuna integration supports only optuna.storages.RDBStorage "
                "in distributed mode"
            )

    def report_metrics(self, validation_cycle: int) -> None:
        """
        Report metrics to the trial.

        This function should be called from the model once the corresponding metrics have
        been saved into the :py:attr:`~helios.model.model.Model.metrics` table.

        Example:
            .. code-block:: python

                import helios.model as hlm
                import helios.plugins.optuna as hlpo

                class MyModel(hlm.Model):
                    ...
                    def on_validation_end(self, validation_cycle: int) -> None:
                        # Compute metrics
                        self.metrics["accuracy"] = 10

                        plugin = self.trainer.plugins[hlpo.OptunaPlugin.plugin_id]
                        assert isinstance(plugin hlpo.OptunaPlugin)
                        plugin.report_metrics(validation_cycle)

        .. note::
            In distributed training, only rank 0 will report the metrics to the trial.

        Args:
            validation_cycle: the current validation cycle.
        """
        model = self.trainer.model
        if not model.metrics or self._metric_name not in model.metrics:
            return

        if self.rank == 0:
            self.trial.report(model.metrics[self._metric_name], validation_cycle)
        self._last_cycle = validation_cycle

    def should_training_stop(self) -> bool:
        """
        Handle trial pruning.

        Returns:
            True if the trial should be pruned, false otherwise.
        """
        should_stop = False
        if self.rank == 0:
            should_stop = self.trial.should_prune()

        # Sync the value across all processes (if using distributed training).
        if self.is_distributed:
            t = dist.all_reduce_tensors(torch.tensor(should_stop).to(self.device))
            should_stop = t.item()  # type: ignore[assignment]

        if should_stop and self.rank == 0:
            self.trial.set_user_attr(_PRUNED_KEY, True)
            self.trial.set_user_attr(_CYCLE_KEY, self._last_cycle)

        return should_stop

    def on_training_end(self) -> None:
        """
        Clean-up on training end.

        If training is non-distributed and the trial was pruned, then this function will
        do the following:

        #. Call :py:meth:`~helios.model.model.Model.on_training_end` to ensure metrics are
           correctly logged (if using).
        #. Raise :py:exc:`optuna.TrialPruned` exception to signal the trial was pruned.

        If training is distributed, this function does nothing.

        Raises:
            TrialPruned: if the trial was pruned.
        """
        if not self.is_distributed and self.trial.should_prune():
            self.trainer.model.on_training_end()
            raise optuna.TrialPruned(f"Pruned on validation cycle {self._last_cycle}")

    def check_pruned(self) -> None:
        """
        Ensure pruned distributed trials are correctly handled.

        Due to the way distributed training works, we can't raise an exception within the
        distributed processes, so we have to do it after we return to the main process.
        If the trial was pruned, this function will raise :py:exc:`optuna.TrialPruned`. If
        distributed training wasn't used, this function does nothing.

        .. warning::
            You *must* ensure this function is called after
            :py:meth:`~helios.trainer.Trainer.fit` to ensure pruning works correctly.

        Raises:
            TrialPruned: if the trial was pruned.
        """
        trial_id = self.trial._trial_id
        study = self.trial.study
        trial = study._storage._backend.get_trial(trial_id)  # type: ignore[attr-defined]
        is_pruned = trial.user_attrs.get(_PRUNED_KEY)
        val_cycle = trial.user_attrs.get(_CYCLE_KEY)

        if is_pruned is None or val_cycle is None:
            return

        if is_pruned:
            raise optuna.TrialPruned(f"Pruned on validation cycle {val_cycle}")

    def state_dict(self) -> dict[str, typing.Any]:
        """
        Get the state of the current trial.

        This will return the parameters to be optimised for the current trial.

        Returns:
            The parameters of the trial.
        """
        return self._trial.params



from __future__ import annotations

import dataclasses as dc
import enum
import itertools
import os
import pathlib
import re
import time
import typing

import packaging.version as pv
import torch
import torch.multiprocessing as mp
import torch.utils.data as tud
import tqdm

import helios.model as hlm
import helios.plugins as hlp
from helios import core, data
from helios.core import distributed as dist
from helios.core import logging, rng
from helios.data.samplers import ResumableSamplerType

from ._version import __version__


class TrainingUnit(enum.Enum):
    """Defines the types of units for training steps."""

    ITERATION = 0
    EPOCH = 1

    @classmethod
    def from_str(cls, label: str) -> TrainingUnit:
        """
        Convert the given string to the corresponding enum value.

        Must be one of "iteration" or "epoch".

        Args:
            label: the label to convert.

        Returns:
            The corresponding value.

        Raises:
            ValueError: if the given value is not one of "iteration" or "epoch".
        """
        if label == "iteration":
            return cls.ITERATION
        if label == "epoch":
            return cls.EPOCH

        raise ValueError(
            "invalid training unit. Expected one of 'iteration' or 'epoch' but "
            f"received '{label}'"
        )


class _TrainerMode(enum.Enum):
    """
    Defines the types of actions that the trainer can do.

    These values are used internally to figure out which function should be invoked from
    the distributed handler.
    """

    TRAIN = 0
    TEST = 1


@dc.dataclass
class _DistributedErrorState:
    """
    Holds the state necessary to correctly handle errors in distributed training.

    This class is pushed into the internal error handling queue and used by the main
    process when an error occurs to ensure the exception is handled correctly.

    Args:
        log_path: the path used by the root logger (if using).
    """

    log_path: pathlib.Path | None = None


@dc.dataclass
class TrainingState:
    """
    The training state.

    Args:
        current_iteration: the current iteration number. Note that this refers to
            the iteration in which gradients are updated. This may or may not be equal to
            the :py:attr:`global_iteration` count.
        global_iteration: the total iteration count.
        global_epoch: the total epoch count.
        validation_cycles: the number of validation cycles.
        dataset_iter: the current batch index of the dataset. This is reset every epoch.
        early_stop_count: the current number of validation cycles for early stop.
        average_iter_time: average time per iteration.
        running_iter: iteration count in the current validation cycle. Useful for
            computing running averages of loss functions.
    """

    current_iteration: int = 0
    global_iteration: int = 0
    global_epoch: int = 0
    validation_cycles: int = 0
    dataset_iter: int = 0
    early_stop_count: int = 0
    average_iter_time: float = 0
    running_iter: int = 0

    dict = dc.asdict


def get_trainer_safe_types_for_load() -> list[
    typing.Callable | tuple[typing.Callable, str]
]:
    """
    Return the list of safe types for loading needed by the trainer.

    Returns:
        The list of types that need to be registered for safe loading.
    """
    return [TrainingState, pathlib.PosixPath, pathlib.WindowsPath]


def register_trainer_types_for_safe_load() -> None:
    """Register trainer types for safe loading."""
    torch.serialization.add_safe_globals(get_trainer_safe_types_for_load())


def find_last_checkpoint(root: pathlib.Path | None) -> pathlib.Path | None:
    """
    Find the last saved checkpoint (if available).

    The function assumes that checkpoint names contain ``epoch_<epoch>`` and
    ``iter_<iter>`` in the name, in which case it will return the path to the checkpoint
    with the highest epoch and/or iteration count.

    Args:
        root: the path where the checkpoints are stored.

    Returns:
        The path to the last checkpoint (if any).
    """
    if root is None:
        return None

    epoch = 0
    ite = 0
    last_chkpt = None
    regexp = re.compile(r"epoch_\d+_iter_\d+")
    for path in root.glob("*.pth"):
        if not regexp.search(path.stem):
            continue

        elems = str(path.stem).split("_")

        idx = elems.index("epoch") + 1
        e = int(elems[idx])

        idx = elems.index("iter") + 1
        i = int(elems[idx])

        if e > epoch or i > ite:
            epoch = e
            ite = i
            last_chkpt = path

    return last_chkpt


def _spawn_handler(
    rank: int,
    world_size: int,
    trainer: Trainer,
    datamodule: data.DataModule,
    model: hlm.Model,
    mode: _TrainerMode,
    queue: mp.Queue,
    error_queue: mp.Queue,
) -> None:
    """
    Spawn handler for distributed training.

    Args:
        rank: the rank id of the current process within the work group.
        world_size: number of processes in the work group.
        trainer: the trainer to use.
        datamodule: the datamodule to use.
        model: the model to use.
        mode: determines which operation needs to be performed.
    """
    dist.init_dist(rank=rank, world_size=world_size)

    trainer.model = model
    trainer.datamodule = datamodule
    trainer.rank = rank
    trainer.local_rank = rank
    trainer.queue = queue
    trainer._distributed_error_queue = error_queue  # noqa: SLF001

    try:
        if mode == _TrainerMode.TRAIN:
            trainer._train()  # noqa: SLF001
        elif mode == _TrainerMode.TEST:
            trainer._test()  # noqa: SLF001
    except Exception as e:
        dist.shutdown_dist()
        raise e

    dist.safe_barrier()
    dist.shutdown_dist()


class Trainer:
    """
    Automates the training, validation, and testing code.

    The trainer handles all of the required steps to setup the correct environment
    (including handling distributed training), the training/validation/testing loops, and
    any clean up afterwards.

    Args:
        run_name: name of the current run. Defaults to empty.
        train_unit: the unit used for training. Defaults to
            :py:attr:`TrainingUnit.ITERATION`.
        total_steps: the total number of steps to train for. Defaults to 0.
        valid_frequency: (optional) frequency with which to perform validation.
        chkpt_frequency: (optional) frequency with which to save checkpoints.
        print_frequency: (optional) frequency with which to log.
        accumulation_steps: number of steps for gradient accumulation. Defaults to 1.
        enable_cudnn_benchmark: enable/disable CuDNN benchmark. Defaults to false.
        enable_deterministic: enable/disable PyTorch deterministic. Defaults to false.
        early_stop_cycles: (optional) number of cycles after which training will stop if
            no improvement is seen during validation.
        use_cpu: (optional) if true, CPU will be used.
        gpus: (optional) IDs of GPUs to use.
        random_seed: (optional) the seed to use for RNGs.
        enable_tensorboard: enable/disable Tensorboard logging. Defaults to false.
        enable_file_logging: enable/disable file logging. Defaults to false.
        capture_warnings: enable/disable warnings capture in the log. Defaults to false.
        enable_progress_bar: enable/disable the progress bar(s). Defaults to false.
        chkpt_root: (optional) root folder in which checkpoints will be placed.
        log_path: (optional) root folder in which logs will be saved.
        run_path: (optional) root folder in which Tensorboard runs will be saved.
        src_root: (optional) root folder where the code is located. This is used to
            automatically populate the registries using
            :py:func:`~helios.core.utils.update_all_registries`.
        import_prefix: prefix to use when importing modules. See
            :py:func:`~helios.core.utils.update_all_registries` for details.
        print_banner: if true, the Helios banner with system info will be printed.
            Defaults to true.
    """

    def __init__(
        self,
        run_name: str = "",
        train_unit: TrainingUnit | str = TrainingUnit.EPOCH,
        total_steps: int | float = 0,
        valid_frequency: int | None = None,
        chkpt_frequency: int | None = None,
        print_frequency: int | None = None,
        accumulation_steps: int = 1,
        enable_cudnn_benchmark: bool = False,
        enable_deterministic: bool = False,
        early_stop_cycles: int | None = None,
        use_cpu: bool | None = None,
        gpus: list[int] | None = None,
        random_seed: int | None = None,
        enable_tensorboard: bool = False,
        enable_file_logging: bool = False,
        capture_warnings: bool = False,
        enable_progress_bar: bool = False,
        chkpt_root: pathlib.Path | None = None,
        log_path: pathlib.Path | None = None,
        run_path: pathlib.Path | None = None,
        src_root: pathlib.Path | None = None,
        import_prefix: str = "",
        print_banner: bool = True,
    ):
        """Create the trainer."""
        self._model: hlm.Model | None = None
        self._datamodule: data.DataModule | None = None
        self._local_rank: int = 0
        self._rank: int = 0

        self._use_cpu: bool = False
        self._device: torch.device | None = None
        self._map_loc: str | dict[str, str] = ""
        self._gpu_ids: list[int] = [] if gpus is None else gpus
        self._active_gpu: int = 0
        self._is_distributed: bool = False
        self._is_torchrun: bool = dist.is_using_torchrun()

        if isinstance(train_unit, str):
            train_unit = TrainingUnit.from_str(train_unit)

        self._train_unit = train_unit
        self._total_steps = total_steps
        self._accumulation_steps = accumulation_steps
        self._valid_frequency = valid_frequency
        self._chkpt_frequency = chkpt_frequency
        self._print_frequency = print_frequency
        self._enable_cudnn_benchmark = enable_cudnn_benchmark
        self._enable_deterministic = enable_deterministic
        self._early_stop_cycles = early_stop_cycles
        self._enable_tensorboard = enable_tensorboard
        self._capture_warnings = capture_warnings
        self._enable_file_logging = enable_file_logging
        self._random_seed = rng.get_default_seed() if random_seed is None else random_seed
        self._enable_progress_bar = enable_progress_bar

        self._chkpt_root = chkpt_root
        self._log_path = log_path
        self._run_path = run_path

        self._src_root = src_root
        self._import_prefix = import_prefix

        self._run_name = run_name
        self._print_banner = print_banner

        self._train_exceptions: list[type[Exception]] = []
        self._test_exceptions: list[type[Exception]] = []

        self._plugins: dict[str, hlp.Plugin] = {}

        self._queue: mp.Queue | None = None
        self._distributed_error_queue: mp.Queue | None = None

        self._validate_flags(use_cpu)
        self._setup_device_flags(use_cpu)

    @property
    def model(self) -> hlm.Model:
        """Return the model."""
        return core.get_from_optional(self._model)

    @model.setter
    def model(self, model: hlm.Model) -> None:
        self._model = model

    @property
    def datamodule(self) -> data.DataModule:
        """Return the datamodule."""
        return core.get_from_optional(self._datamodule)

    @datamodule.setter
    def datamodule(self, datamodule: data.DataModule) -> None:
        self._datamodule = datamodule

    @property
    def local_rank(self) -> int:
        """Return the local rank of the trainer."""
        return self._local_rank

    @local_rank.setter
    def local_rank(self, r) -> None:
        self._local_rank = r
        if not self._use_cpu:
            self._active_gpu = self._gpu_ids[r]

    @property
    def rank(self) -> int:
        """Return the global rank of the trainer."""
        return self._rank

    @rank.setter
    def rank(self, r) -> None:
        self._rank = r

    @property
    def gpu_ids(self) -> list[int]:
        """Return the list of GPU IDs to use for training."""
        return self._gpu_ids

    @property
    def train_exceptions(self) -> list[type[Exception]]:
        """Return the list of valid exceptions for training."""
        return self._train_exceptions

    @train_exceptions.setter
    def train_exceptions(self, exc: list[type[Exception]]) -> None:
        self._train_exceptions = exc

    @property
    def test_exceptions(self) -> list[type[Exception]]:
        """Return the list of valid exceptions for testing."""
        return self._test_exceptions

    @test_exceptions.setter
    def test_exceptions(self, exc: list[type[Exception]]) -> None:
        self._test_exceptions = exc

    @property
    def plugins(self) -> dict[str, hlp.Plugin]:
        """Return the list of plug-ins."""
        return self._plugins

    @plugins.setter
    def plugins(self, plugs: dict[str, hlp.Plugin]) -> None:
        self._plugins = plugs

    @property
    def queue(self) -> mp.Queue | None:
        """
        Return the multi-processing queue instance.

        .. note::
            If training isn't distributed or if `torchrun`, then `None` is returned
            instead.
        """
        return self._queue

    @queue.setter
    def queue(self, q: mp.Queue) -> None:
        self._queue = q

    def fit(self, model: hlm.Model, datamodule: data.DataModule) -> bool:
        """
        Run the full training routine.

        Args:
            model: the model to run on.
            datamodule: the datamodule to use.

        Returns:
            True if the training process completed successfully, false otherwise.
        """
        try:
            self._launch(model, datamodule, _TrainerMode.TRAIN)
        except Exception as e:
            if not self._handle_exception(e, _TrainerMode.TRAIN):
                raise e
            return False

        return True

    def test(self, model: hlm.Model, datamodule: data.DataModule) -> bool:
        """
        Run the full testing routine.

        Args:
            model: the model to run on.
            datamodule: the datamodule to use.

        Returns:
            True if the training process completed successfully, false otherwise.
        """
        try:
            self._launch(model, datamodule, _TrainerMode.TEST)
        except Exception as e:
            if not self._handle_exception(e, _TrainerMode.TEST):
                raise e
            return False

        return True

    def _handle_exception(self, e: Exception, mode: _TrainerMode) -> bool:
        """
        Exception handler.

        Args:
            e: the raised exception.

        Returns:
            False if the exception should be allowed to continue up the stack. If true,
            the exception has been handled and should not be re-raised.
        """
        exc_list = (
            self._train_exceptions
            if mode == _TrainerMode.TRAIN
            else self._test_exceptions
        )
        if any(isinstance(e, exc) for exc in exc_list):
            logging.close_default_loggers()
            return False

        if logging.is_root_logger_active():
            root_logger = logging.get_root_logger()
            root_logger.exception("error: uncaught exception")
            logging.close_default_loggers()
        return True

    def _configure_env_for_distributed_error_handling(self) -> None:
        assert self._distributed_error_queue is not None
        if self._distributed_error_queue.empty():
            return

        state: _DistributedErrorState = self._distributed_error_queue.get_nowait()
        if self._enable_file_logging and state.log_path is not None:
            logging.create_default_loggers(
                enable_tensorboard=False, capture_warnings=True
            )
            logging.restore_default_loggers(log_path=state.log_path)

    def _push_distributed_error_state(self, state: _DistributedErrorState) -> None:
        if self._distributed_error_queue is None:
            return
        self._distributed_error_queue.put(state)

    def _launch(
        self, model: hlm.Model, datamodule: data.DataModule, mode: _TrainerMode
    ) -> None:
        """
        Launch the function corresponding to the given mode.

        If distributed training is used, this will spawn the processes and call the
        handler.

        Args:
            model: the model to use.
            datamodule: the datamodule to use.
            mode:: the operation to perform.
        """
        datamodule.prepare_data()

        if self._is_distributed and not self._is_torchrun:
            if mp.get_start_method(allow_none=True) is None:
                mp.set_start_method("spawn")
            queue: mp.Queue = mp.Queue()
            error_queue: mp.Queue = mp.Queue()
            world_size = len(self._gpu_ids)
            try:
                mp.spawn(
                    _spawn_handler,
                    args=(
                        world_size,
                        self,
                        datamodule,
                        model,
                        mode,
                        queue,
                        error_queue,
                    ),
                    nprocs=world_size,
                    join=True,
                )
            except Exception as e:
                self._distributed_error_queue = error_queue
                self._configure_env_for_distributed_error_handling()
                raise e

            self.queue = queue
            return

        if self._is_torchrun and self._is_distributed:
            dist.init_dist()

        self.model = model
        self.datamodule = datamodule
        self.rank = dist.get_global_rank()
        self.local_rank = dist.get_local_rank()
        if mode == _TrainerMode.TRAIN:
            self._train()
        elif mode == _TrainerMode.TEST:
            self._test()

        if self._is_torchrun and self._is_distributed:
            dist.shutdown_dist()

        logging.close_default_loggers()

    def _train(self) -> None:
        """
        Train the model.

        This will ensure everything gets correctly initialised as well as select the
        appropriate training loop for the given training unit.
        """
        self._configure_env()
        self._setup_plugins()
        self._setup_datamodule()
        self._setup_model()
        self._prepare_roots()

        chkpt_path = find_last_checkpoint(self._chkpt_root)
        training_state = self._load_checkpoint(chkpt_path)

        log_path = logging.get_root_logger().log_file
        self._push_distributed_error_state(_DistributedErrorState(log_path=log_path))

        self._print_header(chkpt_path)

        self._execute_plugins("on_training_start")
        self.model.on_training_start()
        if self._train_unit == TrainingUnit.ITERATION:
            self._train_on_iteration(training_state)
        else:
            self._train_on_epoch(training_state)

        self._execute_plugins("on_training_end")
        self.model.on_training_end()

        logging.flush_default_loggers()
        logging.close_default_loggers()

        # If we're distributed, ensure that all processes are caught up before we exit.
        dist.safe_barrier()

    def _test(self) -> None:
        """
        Test the model.

        This will ensure everything gets correctly initialised and run the testing loop on
        the dataset.
        It will automatically try to load the last saved checkpoint for testing provided
        there is one. If no checkpoints are available, it is assumed the model is loading
        the correct state internally.
        """
        self._configure_env()
        self._setup_plugins()
        self._setup_datamodule()
        self._setup_model(fast_init=True)
        self._prepare_roots(mkdir=False)

        chkpt_path: pathlib.Path | None = None
        loaded: bool = False
        if self._chkpt_root is not None:
            chkpt_path = find_last_checkpoint(core.get_from_optional(self._chkpt_root))
            loaded = (
                self._load_checkpoint(chkpt_path, skip_rng=True, model_fast_init=True)
                != TrainingState()
            )

        # We failed to load the last checkpoint, so tell the model to load its state.
        if self._chkpt_root is None or not loaded:
            self.model.load_for_testing()

        log_path = logging.get_root_logger().log_file
        self._push_distributed_error_state(_DistributedErrorState(log_path=log_path))

        self._print_header(chkpt_path, for_training=False)

        if self.datamodule.test_dataloader() is None:
            return

        dataloader: tud.DataLoader
        sampler: ResumableSamplerType
        dataloader, sampler = core.get_from_optional(self.datamodule.test_dataloader())

        self._execute_plugins("on_testing_start")
        self.model.on_testing_start()

        enable_progress_bar = self._enable_progress_bar
        pbar_disabled = (
            self._is_distributed and self.rank != 0
        ) or not enable_progress_bar
        pbar = tqdm.tqdm(
            total=len(dataloader),
            desc="Testing",
            unit="it",
            disable=pbar_disabled,
            leave=False,
        )

        with core.cuda.DisableCuDNNBenchmarkContext():
            self.model.eval()
            with torch.no_grad():
                for idx, batch in enumerate(dataloader):
                    batch = self._plugins_process_batch("testing", batch, step=idx)
                    self.model.on_testing_batch_start(idx)
                    self.model.test_step(batch, idx)
                    self.model.on_testing_batch_end(idx)
                    pbar.update()

            self._execute_plugins("on_testing_end")
            self.model.on_testing_end()

        dist.safe_barrier()

    def _configure_env(self) -> None:
        """
        Configure the training environment.

        This will seed the RNGs as well as setup any CUDA state (if using). It will also
        set all of the registries provided the source root is not None. This is to prevent
        the registries from being empty if distributed training is launched through spawn
        (note that ``torchrun`` doesn't have this problem).
        """
        safe_types = get_trainer_safe_types_for_load()
        safe_types.extend(self.model.types_for_safe_load())
        torch.serialization.add_safe_globals(safe_types)
        rng.seed_rngs(self._random_seed)
        torch.use_deterministic_algorithms(self._enable_deterministic)

        if not self._use_cpu:
            self._device = torch.device(f"cuda:{self._active_gpu}")
            self._map_loc = {"cuda:0": f"cuda:{self._active_gpu}"}
            torch.backends.cudnn.benchmark = self._enable_cudnn_benchmark
            torch.cuda.set_device(self._device)

        logging.create_default_loggers(
            self._enable_tensorboard, capture_warnings=self._capture_warnings
        )

        if self._src_root is not None:
            core.update_all_registries(
                self._src_root, recurse=True, import_prefix=self._import_prefix
            )

    def _setup_datamodule(self) -> None:
        """Finish setting up the datamodule."""
        self.datamodule.is_distributed = self._is_distributed
        self.datamodule.trainer = self
        self.datamodule.setup()

    def _setup_model(self, fast_init: bool = False) -> None:
        """
        Finish setting up the model.

        Args:
            fast_init: whether the model should setup its full state or not.
        """
        self.model.map_loc = self._map_loc
        self.model.is_distributed = self._is_distributed
        self.model.device = core.get_from_optional(self._device)
        self.model.rank = self.local_rank
        self.model.trainer = self
        self.model.setup(fast_init)

    def _setup_plugins(self) -> None:
        """Finish setting up the plug-ins."""
        for plugin in self._plugins.values():
            plugin.is_distributed = self._is_distributed
            plugin.map_loc = self._map_loc
            plugin.device = core.get_from_optional(self._device)
            plugin.rank = self.local_rank
            plugin.trainer = self
            plugin.setup()

    def _prepare_roots(self, mkdir=True) -> None:
        """Prepare the training roots."""
        name = self.model.save_name
        self._chkpt_root = (
            self._chkpt_root / name if self._chkpt_root is not None else None
        )
        if mkdir:
            if self._chkpt_root is not None:
                self._chkpt_root.mkdir(parents=True, exist_ok=True)

            if self._log_path is not None:
                self._log_path.mkdir(parents=True, exist_ok=True)
            if self._run_path is not None:
                self._run_path.mkdir(parents=True, exist_ok=True)

    def _print_header(
        self, chkpt_path: pathlib.Path | None, for_training: bool = True
    ) -> None:
        """Print the Helios header with system info to the logs."""
        root_logger = logging.get_root_logger()
        model = core.get_from_optional(self._model)
        banner = model.append_to_banner(core.get_env_info_str())

        if self._print_banner:
            dist.global_print(banner)

        if for_training:
            if chkpt_path is not None:
                msg = f"Resuming training from checkpoint {str(chkpt_path)}"
                root_logger.info(msg)
                dist.global_print(f"{msg}\n")
            elif self._print_banner:
                root_logger.info(banner)
        else:
            root_logger.info(core.get_env_info_str())
            msg = (
                f"Testing using checkpoint {str(chkpt_path)}"
                if chkpt_path is not None
                else "Testing from loaded model"
            )
            root_logger.info(msg)
            dist.global_print(f"{msg}\n")

    def _validate_flags(self, use_cpu: bool | None):
        """Ensure that all the settings and flags are valid."""
        if isinstance(self._total_steps, float) and self._total_steps != float("inf"):
            raise ValueError(
                "error: expected 'total_steps' to be of type 'int' or 'infinity', but "
                f"received {self._total_steps}"
            )

        if use_cpu is not None and use_cpu and len(self._gpu_ids) > 0:
            raise ValueError("error: cannot request CPU and GPU training")

        if self._chkpt_frequency is not None and self._chkpt_frequency == 0:
            raise ValueError("error: checkpoint frequency must be greater than 0 or None")

        if self._print_frequency is not None and self._print_frequency == 0:
            raise ValueError("error: print frequency must be greater than 0 or None")

        if self._valid_frequency is not None and self._valid_frequency == 0:
            raise ValueError("error: valid frequency must be greater than 0 or None")

        if self._enable_deterministic and self._enable_cudnn_benchmark:
            raise ValueError(
                "error: CUDA benchmark and deterministic flags are mutually exclusive"
            )

        if self._total_steps == float("inf") and self._early_stop_cycles == 0:
            raise ValueError(
                f"error: given 'total_steps' with value {self._total_steps}, "
                "'early_stop_cycles' must be non-zero"
            )

        if self._enable_tensorboard:
            if self._run_path is None:
                raise ValueError(
                    "error: Tensorboard requested but no run directory was given"
                )

            if self._run_path.exists() and not self._run_path.is_dir():
                raise ValueError("error: run path must be a directory")

        if self._enable_file_logging:
            if self._log_path is None:
                raise ValueError(
                    "error: file logging requested but no log directory was given"
                )

            if self._log_path.exists() and not self._log_path.is_dir():
                raise ValueError("error: log path must be a directory")

        if self._src_root is not None and not self._src_root.is_dir():
            raise ValueError("error: source root must be a directory")

    def _setup_device_flags(self, use_cpu: bool | None):
        """
        Configure the device state.

        If the CPU is being used, this will automatically set the correct settings. If the
        GPU will be used, then it will only verify that the GPU IDs are correct. The
        remaining state will be set afterwards.
        The ``use_cpu`` flag is used to determine whether the CPU will be used for
        training. If it is ``None``, then the value is determined by whether CUDA is
        available.

        Args:
            use_cpu: whether to use the CPU or not.
        """
        if use_cpu is None:
            use_cpu = not torch.cuda.is_available()

        if use_cpu:
            self._use_cpu = True
            self._device = torch.device("cpu")
            self._map_loc = {"cuda:0": "cpu"}
            self._gpu_ids = []
            self._is_distributed = False
            return

        if not torch.cuda.is_available():
            raise RuntimeError(
                "error: CUDA usage is requested, but CUDA is not available"
            )

        # At this point we know that CUDA exists and that we're supposed to use it. For
        # now, just verify that the GPU IDs are valid, but don't set the device or the map
        # location. Those need to be set after we launch distributed training (if using)
        # to ensure they get set to the correct thing.

        valid_ids = list(range(torch.cuda.device_count()))
        if len(self._gpu_ids) == 0:
            self._gpu_ids = valid_ids

        if len(self._gpu_ids) > len(valid_ids):
            raise ValueError(
                f"error: expected a maximum of {len(valid_ids)} GPU IDs but "
                f"received {len(self._gpu_ids)}"
            )

        for gpu_id in self._gpu_ids:
            if gpu_id not in valid_ids:
                raise ValueError(f"error: {gpu_id} is not a valid GPU")

        self._is_distributed = (
            len(self._gpu_ids) > 1
            if not self._is_torchrun
            else int(os.environ["WORLD_SIZE"]) > 1
        )

    def _validate_state_dict(self, state_dict: dict[str, typing.Any]) -> bool:
        """
        Ensure the state table is valid.

        This is to handle the eventuality that someone tries to load a checkpoint that
        wasn't generated by the trainer. It checks that all the required keys appear, and
        it also does a version check. Mainly: we guarantee support for checkpoints
        generated in versions less than or equal to the current version.

        Args:
            state_dict: the state dictionary.

        Returns:
            True if the checkpoint is valid, false otherwise.
        """
        required_keys = ("version", "training_state", "model", "rng")
        if not all(key in state_dict for key in required_keys):
            return False

        # Now check the version to see if it's compatible with us.
        cur_ver = pv.Version(__version__)
        chkpt_ver = pv.Version(state_dict["version"])
        return chkpt_ver <= cur_ver

    def _save_checkpoint(self, state: TrainingState) -> None:
        """
        Save the current training state to a checkpoint.

        This will automatically save the training state, RNG state, as well as the model
        state.

        Args:
            state: the current training state.
        """
        chkpt_root = core.get_from_optional(self._chkpt_root)

        epoch = state.global_epoch
        ite = state.current_iteration
        filename = f"{self.model.save_name}_epoch_{epoch}_iter_{ite}"
        filename = self.model.append_metadata_to_chkpt_name(filename)
        filename += ".pth"

        state_dict: dict[str, typing.Any] = {}
        state_dict["version"] = __version__
        state_dict["training_state"] = state
        state_dict["model"] = self.model.state_dict()
        state_dict["rng"] = rng.get_rng_state_dict()

        if self._enable_file_logging:
            state_dict["log_path"] = logging.get_root_logger().log_file

        if self._enable_tensorboard:
            writer = core.get_from_optional(logging.get_tensorboard_writer())
            state_dict["run_path"] = writer.run_path

        # Add the plug-ins (if using)
        for plug_id, plugin in self._plugins.items():
            state_dict[plug_id] = plugin.state_dict()

        # Safety check.
        assert self._validate_state_dict(state_dict)

        torch.save(state_dict, chkpt_root / filename)

    def _load_checkpoint(
        self,
        chkpt_path: pathlib.Path | None,
        skip_rng: bool = False,
        model_fast_init: bool = False,
    ) -> TrainingState:
        """
        Load the given checkpoint.

        Args:
            chkpt_path: path to the checkpoint to load.
            skip_rng: if True, skip the loading of the RNG states.
            model_fast_init: whether the model should setup its full state or not.

        Returns:
            Returns the loaded training state and ``True`` if the checkpoint was
                loaded successfully. Otherwise it returns an empty training state and
                ``False``.
        """
        if chkpt_path is None:
            logging.setup_default_loggers(self._run_name, self._log_path, self._run_path)
            return TrainingState()

        state_dict = core.safe_torch_load(chkpt_path, map_location=self._map_loc)
        if not self._validate_state_dict(state_dict):
            raise RuntimeError(
                f"error: the checkpoint found at {str(chkpt_path)} is not a "
                "valid checkpoint generated by Helios"
            )
        logging.restore_default_loggers(
            state_dict.get("log_path", None), state_dict.get("run_path", None)
        )
        if not skip_rng:
            rng.load_rng_state_dict(state_dict["rng"])
        self.model.load_state_dict(state_dict["model"], fast_init=model_fast_init)

        for plug_id, plugin in self._plugins.items():
            if plug_id in state_dict:
                plugin.load_state_dict(state_dict[plug_id])

        return state_dict["training_state"]

    def _train_on_iteration(self, state: TrainingState) -> None:
        """
        Run the main loop for iteration-based training.

        Args:
            state: the training state.
        """
        total_steps = self._total_steps
        save_freq = self._chkpt_frequency
        val_freq = self._valid_frequency
        print_freq = self._print_frequency
        accumulation_steps = self._accumulation_steps
        enable_progress_bar = self._enable_progress_bar
        early_stop_cycles = self._early_stop_cycles

        current_iteration_changed: bool = True
        training_done: bool = False
        root_logger = logging.get_root_logger()
        iter_timer = core.AverageTimer()

        pbar_disabled = (
            self._is_distributed and self.rank != 0
        ) or not enable_progress_bar
        pbar = tqdm.tqdm(
            total=total_steps if total_steps != float("inf") else None,
            desc="Training iterations",
            unit="it",
            disable=pbar_disabled,
            initial=state.current_iteration,
        )

        dataloader: tud.DataLoader
        sampler: ResumableSamplerType
        dataloader, sampler = core.get_from_optional(self.datamodule.train_dataloader())

        sampler.start_iter = state.dataset_iter
        self.model.train()

        for epoch in itertools.count(start=state.global_epoch):
            if training_done:
                break

            state.global_epoch += 1
            root_logger.info(f"Starting epoch {epoch + 1}")
            sampler.set_epoch(epoch)
            epoch_start = time.time()
            self.model.on_training_epoch_start(state.global_epoch)

            iter_timer.start()
            for batch in dataloader:
                state.global_iteration += 1
                if state.global_iteration % accumulation_steps == 0:
                    state.current_iteration += 1
                    state.running_iter += 1
                    current_iteration_changed = True
                else:
                    current_iteration_changed = False

                batch = self._plugins_process_batch("training", batch, state=state)
                self.model.on_training_batch_start(state)
                self.model.train_step(batch, state)
                iter_timer.record()
                state.average_iter_time = iter_timer.get_average_time()
                self.model.on_training_batch_end(
                    state,
                    should_log=(
                        False
                        if print_freq is None
                        else state.current_iteration % print_freq == 0
                        and current_iteration_changed
                    ),
                )
                # Depending on how fast the iteration loop is, it is possible that the
                # progress bar isn't refreshed every tick, so make sure it gets re-drawn.
                if state.global_iteration % accumulation_steps == 0 and not pbar.update():
                    pbar.refresh()
                state.dataset_iter += 1
                if (
                    self._plugins_should_training_stop()
                    or self.model.should_training_stop()
                ):
                    training_done = True
                    break

                if (
                    val_freq is not None
                    and state.current_iteration % val_freq == 0
                    and current_iteration_changed
                ):
                    self._validate(state.validation_cycles)
                    state.validation_cycles += 1
                    state.running_iter = 0
                    if not self.model.have_metrics_improved():
                        state.early_stop_count += 1
                    else:
                        state.early_stop_count = 0

                if (
                    save_freq is not None
                    and state.current_iteration % save_freq == 0
                    and self.rank == 0
                    and current_iteration_changed
                ):
                    self._save_checkpoint(state)

                if (
                    early_stop_cycles is not None
                    and state.early_stop_count >= early_stop_cycles
                ):
                    training_done = True
                    break

                if (
                    self._plugins_should_training_stop()
                    or self.model.should_training_stop()
                ):
                    training_done = True
                    break

                if state.current_iteration >= total_steps:
                    training_done = True
                    break

            state.dataset_iter = 0
            self.model.on_training_epoch_end(state.global_epoch)

            root_logger.info(
                f"Epoch {epoch + 1} completed in {time.time() - epoch_start:.2f}s"
            )

    def _train_on_epoch(self, state: TrainingState) -> None:
        """
        Run the main loop for epoch-based training.

        Args:
            state: the training state.
        """
        total_steps = self._total_steps
        save_freq = self._chkpt_frequency
        val_freq = self._valid_frequency
        print_freq = self._print_frequency
        enable_progress_bar = self._enable_progress_bar
        early_stop_cycles = self._early_stop_cycles

        training_done: bool = False
        root_logger = logging.get_root_logger()
        iter_timer = core.AverageTimer()

        pbar_disabled = (
            self._is_distributed and self.rank != 0
        ) or not enable_progress_bar
        pbar = tqdm.tqdm(
            total=total_steps if total_steps != float("inf") else None,
            desc="Training epochs",
            unit="epoch",
            disable=pbar_disabled,
        )

        dataloader: tud.DataLoader
        sampler: ResumableSamplerType
        dataloader, sampler = core.get_from_optional(self.datamodule.train_dataloader())

        sampler.start_iter = state.dataset_iter
        self.model.train()
        iterator = (
            range(state.global_epoch, int(total_steps))
            if total_steps != float("inf")
            else itertools.count(start=state.global_epoch)
        )

        for epoch in iterator:
            if training_done:
                break

            state.global_epoch += 1

            if state.global_epoch > total_steps:
                training_done = True
                break

            root_logger.info(f"Starting epoch {epoch + 1}")
            sampler.set_epoch(epoch)
            epoch_start = time.time()
            self.model.on_training_epoch_start(state.global_epoch)
            iter_timer.start()
            with tqdm.tqdm(
                total=len(dataloader),
                desc=f"Epoch {epoch + 1}",
                unit="iter",
                disable=pbar_disabled,
                leave=False,
            ) as ite_pbar:
                for batch in dataloader:
                    state.global_iteration += 1
                    state.current_iteration += 1
                    state.running_iter += 1

                    batch = self._plugins_process_batch("training", batch, state=state)
                    self.model.on_training_batch_start(state)
                    self.model.train_step(batch, state)
                    iter_timer.record()
                    state.average_iter_time = iter_timer.get_average_time()
                    self.model.on_training_batch_end(
                        state,
                        should_log=(
                            False
                            if print_freq is None
                            else state.current_iteration % print_freq == 0
                        ),
                    )
                    state.dataset_iter += 1
                    if not ite_pbar.update():
                        ite_pbar.refresh()

                    if (
                        self._plugins_should_training_stop()
                        or self.model.should_training_stop()
                    ):
                        training_done = True
                        break

            state.dataset_iter = 0

            if val_freq is not None and state.global_epoch % val_freq == 0:
                self._validate(state.validation_cycles)
                state.running_iter = 0
                if not self.model.have_metrics_improved():
                    state.early_stop_count += 1
                else:
                    state.early_stop_count = 0
                state.validation_cycles += 1
            if (
                save_freq is not None
                and state.global_epoch % save_freq == 0
                and self.rank == 0
            ):
                self._save_checkpoint(state)

            self.model.on_training_epoch_end(state.global_epoch)
            root_logger.info(
                f"Epoch {epoch + 1} completed in {time.time() - epoch_start:.2f}s"
            )
            pbar.update()
            if (
                early_stop_cycles is not None
                and state.early_stop_count >= early_stop_cycles
            ):
                training_done = True

            if self._plugins_should_training_stop() or self.model.should_training_stop():
                training_done = True

    def _validate(self, val_cycle: int) -> None:
        """
        Run the validation loop.

        Args:
            val_cycle: the current validation cycle number.
        """
        if self.datamodule.valid_dataloader() is None:
            return

        dataloader: tud.DataLoader
        sampler: ResumableSamplerType
        dataloader, sampler = core.get_from_optional(self.datamodule.valid_dataloader())

        enable_progress_bar = self._enable_progress_bar
        pbar_disabled = (
            self._is_distributed and self.rank != 0
        ) or not enable_progress_bar
        pbar = tqdm.tqdm(
            total=len(dataloader),
            desc="Validation",
            unit="it",
            disable=pbar_disabled,
            leave=False,
        )

        with core.cuda.DisableCuDNNBenchmarkContext():
            self.model.eval()
            self._execute_plugins("on_validation_start", validation_cycle=val_cycle)
            self.model.on_validation_start(val_cycle)
            with torch.no_grad():
                for idx, batch in enumerate(dataloader):
                    batch = self._plugins_process_batch("validation", batch, step=idx)
                    self.model.on_validation_batch_start(idx)
                    self.model.valid_step(batch, idx)
                    self.model.on_validation_batch_end(idx)

                    # Ensure the progress bar is updated in the event that the validation
                    # loop runs faster than the refresh rate of the progress bar.
                    if not pbar.update():
                        pbar.refresh()

            self.model.train()
            self._execute_plugins("on_validation_end", validation_cycle=val_cycle)
            self.model.on_validation_end(val_cycle)

        dist.safe_barrier()

    def _validate_plugins(self) -> None:
        seen_overrides: dict[str, str] = {}
        fields = dc.fields(hlp.UniquePluginOverrides)
        for _, plugin in self._plugins.items():
            for field in fields:
                name = field.name
                if getattr(plugin.unique_overrides, name):
                    if name not in seen_overrides:
                        seen_overrides[name] = str(type(plugin))
                    else:
                        raise ValueError(
                            f"error: override field {name} has already been overridden "
                            f"by {seen_overrides[name]}"
                        )

    def _execute_plugins(self, func_name: str, **kwargs: typing.Any) -> None:
        for plugin in self._plugins.values():
            func = getattr(plugin, func_name)
            func(**kwargs)

    def _plugins_process_batch(
        self, mode: str, batch: typing.Any, **kwargs: typing.Any
    ) -> typing.Any:
        func_name = f"process_{mode}_batch"
        override_name = f"{mode}_batch"

        for plugin in self._plugins.values():
            if getattr(plugin.unique_overrides, override_name):
                return getattr(plugin, func_name)(batch=batch, **kwargs)

        return batch

    def _plugins_should_training_stop(self) -> bool:
        return any(plugin.should_training_stop() for plugin in self._plugins.values())



from __future__ import annotations

import abc
import copy
import dataclasses as dc
import enum
import types
import typing

import torch
import torch.utils.data as tud

from helios import core
from helios.core import rng

if typing.TYPE_CHECKING:
    from ..trainer import Trainer

from .samplers import (
    ResumableDistributedSampler,
    ResumableRandomSampler,
    ResumableSamplerType,
    ResumableSequentialSampler,
)


def _register_default_collate(registry: core.Registry) -> None:
    registry.register(tud.default_collate)


DATASET_REGISTRY = core.Registry("dataset")
"""
Global instance of the registry for datasets.

Example:
    .. code-block:: python

        import helios.data as hld

        # This automatically registers your dataset.
        @hld.DATASET_REGISTRY.register
        class MyDataset:
            ...

        # Alternatively you can manually register a dataset like this:
        hld.DATASET_REGISTRY.register(MyDataset)
"""

COLLATE_FN_REGISTRY = core.Registry("collate_fn")
"""
Global instance of the registry for collate functions.

Example:
    .. code-block:: python

        import helios.data as hld

        # This automatically registers your collate function.
        @hld.COLLATE_FN_REGISTRY
        def my_collate_fn():
            ...

        # Alternatively you can manually register a collate function like this:
        hld.COLLATE_FN_REGISTRY.register(my_collate_fn)
"""

_register_default_collate(COLLATE_FN_REGISTRY)


def create_dataset(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> tud.Dataset:
    """
    Create a dataset of the given type.

    This uses ``DATASET_REGISTRY`` to look-up dataset types, so ensure your datasets
    have been registered before using this function.

    Args:
        type_name: the type of the dataset to create.
        args: positional arguments to pass into the dataset.
        kwargs: keyword arguments to pass into the dataset.

    Returns:
        The constructed dataset.
    """
    return DATASET_REGISTRY.get(type_name)(*args, **kwargs)


def create_collate_fn(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> typing.Callable:
    """
    Create a collate function of the given type.

    This uses ``COLLATE_FN_REGISTRY`` to look-up collate functions, so ensure that your
    functions have been registered before using this function.
    In order to support regular functions as well as callable objects, this function
    behaves as follows:

    #. If ``type_name`` refers to a function-type, then ``args`` and ``kwargs`` are
        ignored.
    #. If ``type_name`` refers to an actual type (i.e. a class), then ``args`` and
        ``kwargs`` are forwarded upon instantiation.

    Args:
        type_name: the type of the function to create.
        args: positional arguments to pass into the collate class
        kwargs: keyword arguments to pass into the collate class

    Returns:
        The function
    """
    fn_type = COLLATE_FN_REGISTRY.get(type_name)
    if isinstance(fn_type, types.FunctionType):
        return fn_type
    assert isinstance(fn_type, type), f"Expected a type but found {type(fn_type)}"
    return fn_type(*args, **kwargs)


class DatasetSplit(enum.Enum):
    """The different dataset splits."""

    TRAIN = 0
    VALID = 1
    TEST = 2

    @staticmethod
    def from_str(label: str) -> DatasetSplit:
        """
        Convert the given string to the corresponding enum value.

        Must be one of "train", "test", or "valid"

        Args:
            label: the label to convert.

        Returns:
            The corresponding enum value.

        Raises:
            ValueError: if the given value is not one of "train", "test", or "valid".
        """
        if label == "train":
            return DatasetSplit.TRAIN
        if label == "valid":
            return DatasetSplit.VALID
        if label == "test":
            return DatasetSplit.TEST

        raise ValueError(
            "invalid dataset split. Expected one of 'train', 'test', or "
            f"'valid', but received '{label}'"
        )


def _seed_worker(worker_id: int) -> None:
    worker_seed = torch.initial_seed() % 2**32
    rng.seed_rngs(worker_seed, skip_torch=True)


def create_dataloader(
    dataset: tud.Dataset,
    random_seed: int = rng.get_default_seed(),
    batch_size: int = 1,
    shuffle: bool = False,
    num_workers: int = 0,
    pin_memory: bool = True,
    drop_last: bool = False,
    debug_mode: bool = False,
    is_distributed: bool = False,
    sampler: ResumableSamplerType | None = None,
    collate_fn: typing.Callable | None = None,
) -> tuple[tud.DataLoader, ResumableSamplerType]:
    """
    Create the dataloader for the given dataset.

    If no sampler is provided, the choice of sampler will be determined based on the
    values of is_distributed and shuffle. Specifically, the following logic is used:

    * If is_distributed, then sampler is
      :py:class:`~helios.data.samplers.ResumableDistributedSampler`.
    * Otherwise, if shuffle then sampler is
      :py:class:`~helios.data.samplers.ResumableRandomSampler`, else
      :py:class:`~helios.data.samplers.ResumableSequentialSampler`.

    You may override this behaviour by providing your own sampler instance.

    .. warning::
        If you provide a custom sampler, then it **must** be derived from one of
        :py:class:`helios.data.samplers.ResumableSampler` or
        :py:class:`helios.data.samplers.ResumableDistributedSampler`.

    Args:
        dataset: the dataset to use.
        random_seed: value to use as seed for the worker processes. Defaults to the value
            returned by :py:func:`~helios.core.rng.get_default_seed`.
        batch_size: number of samplers per batch. Defaults to 1.
        shuffle: if true, samples are randomly shuffled. Defaults to false.
        num_workers: number of worker processes for loading data. Defaults to 0.
        pin_memory: if true, use page-locked device memory. Defaults to true.
        drop_last: if true, remove the final batch. Defaults to false.
        debug_mode: if true, then ``num_workers`` will be set to 0. Defaults to false.
        is_distributed: if true, create the distributed sampler. Defaults to false.
        sampler: (optional) sampler to use.
        collate_fn: (optional) function to merge batches.

    Returns:
        The dataloader and sampler.

    Raises:
        TypeError: if ``sampler`` is not ``None`` and not derived from one of
            :py:class:`~helios.data.samplers.ResumableDistributedSampler` or
            :py:class:`~helios.data.samplers.ResumableSampler`.
    """
    assert len(typing.cast(typing.Sized, dataset))

    if sampler is None:
        if is_distributed:
            sampler = ResumableDistributedSampler(
                dataset, shuffle=shuffle, drop_last=drop_last, seed=random_seed
            )
        else:
            if shuffle:
                sampler = ResumableRandomSampler(
                    dataset,  # type: ignore[arg-type]
                    seed=random_seed,
                    batch_size=batch_size,
                )
            else:
                sampler = ResumableSequentialSampler(
                    dataset,  # type: ignore[arg-type]
                    batch_size=batch_size,
                )

    elif not isinstance(sampler, typing.get_args(ResumableSamplerType)):
        raise TypeError(
            "error: expected sampler to derive from one of ResumableSampler or "
            f"ResumableDistributedSampler, but received {type(sampler)}"
        )

    return (
        tud.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers if not debug_mode else 0,
            pin_memory=pin_memory,
            drop_last=drop_last,
            sampler=sampler,
            worker_init_fn=_seed_worker,
            collate_fn=collate_fn,
        ),
        sampler,
    )


@dc.dataclass
class DataLoaderParams:
    """
    Params used to create the dataloader object.

    Args:
        random_seed: value to use as seed for the worker processes.
        batch_size: number of samplers per batch.
        shuffle: if true, samples are randomly shuffled.
        num_workers: number of worker processes for loading data.
        pin_memory: if true, use page-locked device memory.
        drop_last: if true, remove the final batch.
        debug_mode: if true, set number of workers to 0.
        is_distributed: if true, create the distributed sampler.
        sampler: (optional) sampler to use.
        collate_fn: (optional) function to merge batches.
    """

    random_seed: int = rng.get_default_seed()
    batch_size: int = 1
    shuffle: bool = False
    num_workers: int = 0
    pin_memory: bool = True
    drop_last: bool = False
    debug_mode: bool = False
    is_distributed: bool | None = None
    sampler: ResumableSamplerType | None = None
    collate_fn: typing.Callable | None = None

    def to_dict(self) -> dict[str, typing.Any]:
        """Convert the params object to a dictionary using shallow copies."""
        return {field.name: getattr(self, field.name) for field in dc.fields(self)}

    @classmethod
    def from_dict(cls, table: dict[str, typing.Any]):
        """Create a new params object from the given table."""
        params = cls()
        keys = [field.name for field in dc.fields(params)]

        # Skip the sampler key, since that one needs to be created differently.
        for key in keys:
            if key in table and key != "sampler":
                setattr(params, key, table[key])

        return params


@dc.dataclass
class Dataset:
    """
    The dataset and corresponding data loader params.

    Args:
        dataset: the dataset.
        params: the data loader params.
    """

    dataset: tud.Dataset
    params: DataLoaderParams

    def dict(self) -> dict[str, typing.Any]:
        """Convert to a dictionary."""
        res = self.params.to_dict()
        res["dataset"] = self.dataset
        return res


class DataModule(abc.ABC):
    """
    Base class that groups together the creation of the main training datasets.

    The use of this class is to standardize the way datasets and their respective
    dataloaders are created, thereby allowing consistent settings across models.

    Example:
        .. code-block:: python

            from torchvision.datasets import CIFAR10
            from helios import data

            class MyDataModule(data.DataModule):
                def prepare_data(self) -> None:
                    # Use this function to prepare the data for your datasets. This will
                    # be called before the distributed processes are created (if using)
                    # so you should not set any state here.
                    CIFAR10(download=True) # download the dataset only.

                def setup(self) -> None:
                    # Create the training dataset using a DataLoaderParams instance. Note
                    # that you MUST assign it to self._train_dataset.
                    self._train_dataset = self._create_dataset(CIFAR10(train=True),
                                                               DataLoaderParams(...))

                    # It is also possible to create a dataset using a table of key-value
                    # pairs that was loaded from a config file or manually created. Let's
                    # use one to create the validation split:
                    settings = {"batch_size": 1, ...}

                    # We can now use it to assign to self._valid_dataset like this:
                    self._valid_dataset = self._create_dataset(
                            CIFAR10(train=False), settings)

                    # Finally, if you need a testing split, you can create it like this:
                    self._test_dataset = self._create_dataset(
                            CIFAR10(train=False), settings)

                def teardown(self) -> None:
                    # Use this function to clean up any state. It will be called after
                    # training is done.
    """

    def __init__(self) -> None:
        """Create the data module."""
        self._is_distributed: bool = False
        self._train_dataset: Dataset | None = None
        self._valid_dataset: Dataset | None = None
        self._test_dataset: Dataset | None = None
        self._trainer: Trainer | None = None

    @property
    def is_distributed(self) -> bool:
        """Flag controlling whether distributed training is being used or not."""
        return self._is_distributed

    @is_distributed.setter
    def is_distributed(self, val: bool) -> None:
        self._is_distributed = val

    @property
    def trainer(self) -> Trainer:
        """Reference to the trainer."""
        return core.get_from_optional(self._trainer)

    @trainer.setter
    def trainer(self, t) -> None:
        self._trainer = t

    @property
    def train_dataset(self) -> tud.Dataset | None:
        """The training dataset (if available)."""
        if self._train_dataset is not None:
            return self._train_dataset.dataset
        return None

    @property
    def valid_dataset(self) -> tud.Dataset | None:
        """The validation dataset (if available)."""
        if self._valid_dataset is not None:
            return self._valid_dataset.dataset
        return None

    @property
    def test_dataset(self) -> tud.Dataset | None:
        """The testing dataset (if available)."""
        if self._test_dataset is not None:
            return self._test_dataset.dataset
        return None

    def prepare_data(self) -> None:  # noqa: B027
        """
        Prepare data for training.

        This can include downloading datasets, preparing caches, or streaming them from
        external services. This function will be called on the primary process when using
        distributed training (will be called prior to initialization of the processes) so
        don't store any state here.
        """

    @abc.abstractmethod
    def setup(self) -> None:
        """Construct all required datasets."""

    def train_dataloader(self) -> tuple[tud.DataLoader, ResumableSamplerType] | None:
        """Create the train dataloader (if available)."""
        if self._train_dataset is None:
            return None
        return self._create_dataloader(self._train_dataset)

    def valid_dataloader(self) -> tuple[tud.DataLoader, ResumableSamplerType] | None:
        """Create the valid dataloader (if available)."""
        if self._valid_dataset is None:
            return None
        return self._create_dataloader(self._valid_dataset)

    def test_dataloader(self) -> tuple[tud.DataLoader, ResumableSamplerType] | None:
        """Create the test dataloader (if available)."""
        if self._test_dataset is None:
            return None
        return self._create_dataloader(self._test_dataset)

    def teardown(self) -> None:  # noqa: B027
        """Clean up any state after training is over."""

    def _create_dataset(
        self, dataset: tud.Dataset, params: DataLoaderParams | dict[str, typing.Any]
    ) -> Dataset:
        """
        Create a dataset object from a Torch dataset and a params.

        This is a convenience function to help you create the dataset objects. The params
        object can either be the fully filled in DataLoaderParams instance, or it can be a
        dict containing all of the necessary settings. If a dict is passed in, it will be
        used to populate the corresponding DataLoaderParams object.

        Args:
            dataset: the dataset.
            params: either a params object or a dict.

        Returns:
            The new dataset object.
        """
        return Dataset(
            dataset,
            copy.deepcopy(params)
            if isinstance(params, DataLoaderParams)
            else DataLoaderParams.from_dict(params),
        )

    def _create_dataloader(
        self, dataset: Dataset
    ) -> tuple[tud.DataLoader, ResumableSamplerType]:
        # Only override the distributed flag if it hasn't been set by the user.
        if dataset.params.is_distributed is None:
            dataset.params.is_distributed = self._is_distributed
        return create_dataloader(**dataset.dict())
